import pathlib
import shutil
from typing import Optional, Sequence, Type, Union

from .. import io
from .documenters import (
    ClassDocumenter,
    FunctionDocumenter,
    MemberDocumenter,
    ModuleDocumenter,
    RootDocumenter,
)
from .nodes import ModuleNode, PackageNode


class APIBuilder(object):
    """
    A builder of reStructuredText API documentation for Python packages.

    ::

        >>> import uqbar
        >>> import tempfile
        >>> with tempfile.TemporaryDirectory() as target_directory:
        ...     with uqbar.io.DirectoryChange(target_directory):
        ...         builder = uqbar.apis.APIBuilder(
        ...             initial_source_paths=uqbar.apis.__path__,
        ...             target_directory=target_directory,
        ...             )
        ...         visited_paths = builder()
        ...
        wrote index.rst
        wrote uqbar/index.rst
        wrote uqbar/apis/index.rst
        wrote uqbar/apis/builders.rst
        wrote uqbar/apis/documenters.rst
        wrote uqbar/apis/dummy.rst
        wrote uqbar/apis/graphs.rst
        wrote uqbar/apis/nodes.rst
        wrote uqbar/apis/summarizers.rst

    :param initial_source_paths: a list of paths to scan for Python sources
    :param target_directory: where to write reStructuredText output
    :param document_private_members: whether to document private module members
    :param document_private_modules: whether to document private modules
    :param member_documenter_classes: a list of
        :py:class:`~uqbar.apis.MemberDocumenter` subclasses, defining what classes
        to use to identify and document module members
    :param module_documenter_class: a :py:class:`~uqbar.apis.ModuleDocumenter` subclass
    :param root_documenter_class: a :py:class:`~uqbar.apis.RootDocumenter` subclass
    """

    ### INITIALIZER ###

    def __init__(
        self,
        initial_source_paths: Sequence[Union[str, pathlib.Path]],
        target_directory: Union[str, pathlib.Path],
        document_empty_modules: bool = True,
        document_private_members: bool = False,
        document_private_modules: bool = False,
        member_documenter_classes: Optional[Sequence[Type[MemberDocumenter]]] = None,
        module_documenter_class: Optional[Type[ModuleDocumenter]] = None,
        omit_root: bool = False,
        root_documenter_class: Optional[Type[RootDocumenter]] = None,
        title: str = "API",
        logger_func=None,
    ) -> None:
        assert initial_source_paths
        assert target_directory
        self._initial_source_paths = frozenset(
            pathlib.Path(_).resolve().absolute() for _ in initial_source_paths
        )
        self._target_directory = pathlib.Path(target_directory).absolute()
        self._document_private_members = bool(document_private_members)
        self._document_private_modules = bool(document_private_modules)
        self._document_empty_modules = bool(document_empty_modules)
        if member_documenter_classes is None:
            member_documenter_classes = [ClassDocumenter, FunctionDocumenter]
        for _ in member_documenter_classes:
            assert issubclass(_, MemberDocumenter), _
        self._member_documenter_classes = tuple(member_documenter_classes)
        if module_documenter_class is None:
            module_documenter_class = ModuleDocumenter
        assert issubclass(module_documenter_class, ModuleDocumenter)
        self._module_documenter_class = module_documenter_class
        self._omit_root = bool(omit_root)
        if root_documenter_class is None:
            root_documenter_class = RootDocumenter
        assert issubclass(root_documenter_class, RootDocumenter)
        self._root_documenter_class = root_documenter_class
        self._title = title
        self._logger_func = logger_func

    ### SPECIAL METHODS ###

    def __call__(self):
        """
        Generate documentation.
        """
        import uqbar.apis

        # Make sure target directory exists.
        if not self._target_directory.exists():
            self._target_directory.mkdir(parents=True)
        self._target_directory = self._target_directory.resolve()
        # What files to document?
        source_paths = uqbar.apis.collect_source_paths(self._initial_source_paths)
        # Build the node tree
        node_tree = self.build_node_tree(source_paths)
        # Get the documenters, in depth-first order.
        documenters = self.collect_module_documenters(node_tree)
        # Execute each documenter, writing to the target directory.
        visited_paths = self.write(documenters)
        # Iterate the target directory, removing un-touched files.
        self.prune(visited_paths)
        return visited_paths

    ### PRIVATE METHODS ###

    def _print(self, message, path):
        cwd = pathlib.Path.cwd()
        if str(path).startswith(str(cwd)):
            path = path.relative_to(cwd)
        message = "{} {}".format(message, path)
        (self._logger_func or print)(message)

    ### PUBLIC METHODS ###

    def build_node_tree(self, source_paths):
        """
        Build a node tree.
        """
        import uqbar.apis

        root = PackageNode()
        # Build node tree, top-down
        for source_path in sorted(
            source_paths, key=lambda x: uqbar.apis.source_path_to_package_path(x)
        ):
            package_path = uqbar.apis.source_path_to_package_path(source_path)
            parts = package_path.split(".")
            if not self.document_private_modules and any(
                part.startswith("_") for part in parts
            ):
                continue
            # Find parent node.
            parent_node = root
            if len(parts) > 1:
                parent_package_path = ".".join(parts[:-1])
                try:
                    parent_node = root[parent_package_path]
                except KeyError:
                    parent_node = root
                try:
                    if parent_node is root:
                        # Backfill missing parent node.
                        grandparent_node = root
                        if len(parts) > 2:
                            grandparent_node = root[
                                parent_package_path.rpartition(".")[0]
                            ]
                        parent_node = PackageNode(name=parent_package_path)
                        grandparent_node.append(parent_node)
                        grandparent_node[:] = sorted(
                            grandparent_node, key=lambda x: x.package_path
                        )
                except KeyError:
                    parent_node = root
            # Create or update child node.
            node_class = ModuleNode
            if source_path.name == "__init__.py":
                node_class = PackageNode
            try:
                # If the child exists, it was previously backfilled.
                child_node = root[package_path]
                child_node.source_path = source_path
            except KeyError:
                # Otherwise it needs to be created and appended to the parent.
                child_node = node_class(name=package_path, source_path=source_path)
                parent_node.append(child_node)
                parent_node[:] = sorted(parent_node, key=lambda x: x.package_path)
        # Build documenters, bottom-up.
        # This allows parent documenters to easily aggregate their children.
        for node in root.depth_first(top_down=False):
            kwargs = dict(
                document_private_members=self.document_private_members,
                member_documenter_classes=self.member_documenter_classes,
            )
            if isinstance(node, ModuleNode):
                node.documenter = self.module_documenter_class(
                    node.package_path, **kwargs
                )
            else:
                # Collect references to child modules and packages.
                node.documenter = self.module_documenter_class(
                    node.package_path,
                    module_documenters=[
                        child.documenter
                        for child in node
                        if child.documenter is not None
                    ],
                    **kwargs,
                )
            if (
                not self.document_empty_modules
                and not node.documenter.module_documenters
                and not node.documenter.member_documenters
            ):
                node.parent.remove(node)
        return root

    def collect_module_documenters(self, root_node):
        if not self.omit_root:
            # Yield root documenter.
            yield self.root_documenter_class(
                module_documenters=[node.documenter for node in root_node],
                title=self._title,
            )
        # Yield module documenters, top-down.
        for node in root_node.depth_first():
            if node is not root_node:
                yield node.documenter

    def write(self, documenters):
        visited_paths = set()
        for documenter in documenters:
            new_documentation = str(documenter)
            path = self._target_directory.joinpath(documenter.documentation_path)
            cwd = pathlib.Path.cwd()
            if str(path).startswith(str(cwd)):
                path = path.relative_to(pathlib.Path.cwd())
            io.write(
                new_documentation, path, verbose=True, logger_func=self._logger_func
            )
            visited_paths.add(path.absolute())
        return visited_paths

    def prune(self, visited_paths):
        generator = io.walk(self._target_directory, top_down=False)
        for root_path, directory_paths, file_paths in generator:
            for file_path in file_paths[:]:
                if file_path.suffix != ".rst":
                    continue
                if file_path not in visited_paths:
                    file_paths.remove(file_path)
                    file_path.unlink()
                    self._print("pruned", file_path)
            if not file_paths and not directory_paths:
                if root_path == self._target_directory:
                    continue
                shutil.rmtree(str(root_path))
                self._print("pruned", root_path)

    ### PUBLIC PROPERTIES ###

    @property
    def document_empty_modules(self):
        return self._document_empty_modules

    @property
    def document_private_members(self):
        return self._document_private_members

    @property
    def document_private_modules(self):
        return self._document_private_modules

    @property
    def initial_source_paths(self):
        return self._initial_source_paths

    @property
    def member_documenter_classes(self):
        return self._member_documenter_classes

    @property
    def module_documenter_class(self):
        return self._module_documenter_class

    @property
    def omit_root(self):
        return self._omit_root

    @property
    def target_directory(self):
        return self._target_directory

    @property
    def root_documenter_class(self):
        return self._root_documenter_class



import collections


class DependencyGraph:
    """
    A dependency graph of hashables.

    TODO: Switch parent/child terminology.

    ::

        >>> import uqbar.containers
        >>> graph = uqbar.containers.DependencyGraph()
        >>> graph.add('A', parent='B')
        >>> graph.add('C', parent='A')
        >>> graph.add('C', parent='B')
        >>> graph.add('D', parent='C')
        >>> graph.add('E')
        >>> graph.add('F', parent='B')
        >>> graph.add('F', parent='C')

    ::

        >>> graph.is_acyclic()
        True

    ::

        >>> list(graph)
        ['D', 'E', 'F', 'C', 'A', 'B']

    ::

        >>> graph.add('A', parent='F')
        >>> graph.is_acyclic()
        False

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_parents_to_children", "_children_to_parents")

    ### INITIALIZER ###

    def __init__(self):
        self._parents_to_children = collections.OrderedDict()
        self._children_to_parents = collections.OrderedDict()

    ### SPECIAL METHODS ###

    def __contains__(self, expr):
        return expr in self._parents_to_children

    def __eq__(self, expr):
        if isinstance(expr, type(self)):
            return self._parents_to_children == expr._parents_to_children

    def __getitem__(self, expr):
        if expr not in self:
            raise ValueError("{!r} not in {}".format(expr, type(self)))
        parents = frozenset(self._children_to_parents[expr])
        children = frozenset(self._parents_to_children[expr])
        return parents, children

    def __iter__(self):
        trellis = self.copy()
        while len(trellis):
            yield trellis.pop()

    def __len__(self):
        return len(self._parents_to_children)

    ### PUBLIC METHODS ###

    def add(self, child, parent=None):
        if parent is not None and parent not in self:
            self.add(parent)
        self._parents_to_children.setdefault(child, [])
        parents = self._children_to_parents.setdefault(child, [])
        if parent is not None:
            if parent not in parents:
                parents.append(parent)
            children = self._parents_to_children[parent]
            if child not in children:
                children.append(child)

    def children(self, expr):
        if expr not in self:
            raise ValueError("{!r} not in {}".format(expr, type(self)))
        return frozenset(self._parents_to_children[expr])

    def copy(self):
        copied = type(self)()
        copied._parents_to_children = self._parents_to_children.copy()
        for parent, children in copied._parents_to_children.items():
            copied._parents_to_children[parent] = list(children)
        copied._children_to_parents = self._children_to_parents.copy()
        for child, parents in copied._children_to_parents.items():
            copied._children_to_parents[child] = list(parents)
        return copied

    def is_acyclic(self):
        trellis = self.copy()
        while len(trellis):
            try:
                trellis.pop()
            except ValueError:
                return False
        return True

    def parents(self, expr):
        if expr not in self:
            raise ValueError("{!r} not in {}".format(expr, type(self)))
        return frozenset(self._children_to_parents[expr])

    def pop(self):
        popped = None
        for parent, children in self._parents_to_children.items():
            if children:
                continue
            popped = parent
            break
        if popped is None:
            raise ValueError("Graph contains cycles.")
        self.remove(popped)
        return popped

    def remove(self, expr):
        if expr not in self:
            raise ValueError("{!r} not in {}".format(expr, type(self)))
        for child in self._parents_to_children.pop(expr):
            self._children_to_parents[child].remove(expr)
        for parent in self._children_to_parents.pop(expr):
            self._parents_to_children[parent].remove(expr)



import collections
import enum
from typing import Generic, TypeVar

import pytest

import uqbar.objects


class MyObject:
    def __init__(self, arg1, arg2, *var_args, foo=None, bar=None, **kwargs):
        self.arg1 = arg1
        self.arg2 = arg2
        self.var_args = var_args
        self.foo = foo
        self.bar = bar
        self.kwargs = kwargs


class Enumeration(enum.IntEnum):
    FOO = 1
    BAR = 2
    BAZ = 3


T = TypeVar("T")


class MyGeneric(Generic[T]):
    def __init__(self, arg: T) -> None:
        self.arg = arg


@pytest.mark.parametrize(
    "object_, expected",
    [
        (
            MyObject("a", "b", "c", "d", foo="x", quux=["y", "z"]),
            (
                collections.OrderedDict([("arg1", "a"), ("arg2", "b")]),
                ["c", "d"],
                {"bar": None, "foo": "x", "quux": ["y", "z"]},
            ),
        ),
        (Enumeration.BAZ, (collections.OrderedDict([("value", 3)]), [], {})),
        (MyGeneric[int](arg=3), (collections.OrderedDict([("arg", 3)]), [], {})),
    ],
)
def test_objects_get_vars(object_, expected):
    actual = uqbar.objects.get_vars(object_)
    assert actual == expected



import sys

import pytest
from docutils.parsers.rst import directives

import uqbar.book.sphinx
from uqbar.book.console import ConsoleError, ConsoleInput, ConsoleOutput
from uqbar.book.extensions import GraphExtension
from uqbar.book.sphinx import UqbarBookDirective
from uqbar.strings import normalize


@pytest.fixture(scope="module", autouse=True)
def register_directives():
    directives.register_directive("book", UqbarBookDirective)
    yield
    directives._directives.pop("book")


source_a = """
::

    >>> string = 'Hello, world!'

::

    >>> for i in range(3):
    ...     string += " {}".format(i)
    ...

::

    >>> print(string)
    Hello, world!
"""

source_b = """
::

    >>> import uqbar.graphs
    >>> g = uqbar.graphs.Graph()
    >>> n1 = uqbar.graphs.Node()
    >>> n2 = uqbar.graphs.Node()
    >>> g.extend([n1, n2])
    >>> e = n1.attach(n2)

::

    >>> print(format(g, "graphviz"))

::

    >>> uqbar.graphs.Grapher(g)()

"""


source_c = """
::

    >>> import uqbar.graphs
    >>> g = uqbar.graphs.Graph()
    >>> n1 = uqbar.graphs.Node()
    >>> g.append(n1)
    >>> for i in range(3):
    ...     n2 = uqbar.graphs.Node()
    ...     g.append(n2)
    ...     e = n1.attach(n2)
    ...     uqbar.graphs.Grapher(g)()
    ...     print(i)
    ...     n1 = n2
    ...

"""

source_d = """
..  book::
    :hide:

    >>> import uqbar.graphs
    >>> g = uqbar.graphs.Graph()
    >>> n1 = uqbar.graphs.Node()
    >>> n2 = uqbar.graphs.Node()
    >>> g.extend([n1, n2])
    >>> e = n1.attach(n2)
    >>> uqbar.graphs.Grapher(g)()

"""


def test_parse_rst_01():
    document = uqbar.book.sphinx.parse_rst(source_a)
    assert normalize(document.pformat()) == normalize(
        """
        <document source="test">
            <literal_block xml:space="preserve">
                >>> string = 'Hello, world!'
            <literal_block xml:space="preserve">
                >>> for i in range(3):
                ...     string += " {}".format(i)
                ...
            <literal_block xml:space="preserve">
                >>> print(string)
                Hello, world!
        """
    )


def test_parse_rst_02():
    document = uqbar.book.sphinx.parse_rst(source_b)
    assert normalize(document.pformat()) == normalize(
        """
        <document source="test">
            <literal_block xml:space="preserve">
                >>> import uqbar.graphs
                >>> g = uqbar.graphs.Graph()
                >>> n1 = uqbar.graphs.Node()
                >>> n2 = uqbar.graphs.Node()
                >>> g.extend([n1, n2])
                >>> e = n1.attach(n2)
            <literal_block xml:space="preserve">
                >>> print(format(g, "graphviz"))
            <literal_block xml:space="preserve">
                >>> uqbar.graphs.Grapher(g)()
        """
    )


def test_collect_literal_blocks_01():
    document = uqbar.book.sphinx.parse_rst(source_a)
    blocks = uqbar.book.sphinx.collect_literal_blocks(document)
    expected = [
        """
        <literal_block xml:space="preserve">
            >>> string = 'Hello, world!'
        """,
        """
        <literal_block xml:space="preserve">
            >>> for i in range(3):
            ...     string += " {}".format(i)
            ...
        """,
        """
        <literal_block xml:space="preserve">
            >>> print(string)
            Hello, world!
        """,
    ]
    actual = [normalize(block.pformat()) for block in blocks]
    expected = [normalize(text) for text in expected]
    assert actual == expected


def test_collect_literal_blocks_02():
    document = uqbar.book.sphinx.parse_rst(source_b)
    blocks = uqbar.book.sphinx.collect_literal_blocks(document)
    expected = [
        """
        <literal_block xml:space="preserve">
            >>> import uqbar.graphs
            >>> g = uqbar.graphs.Graph()
            >>> n1 = uqbar.graphs.Node()
            >>> n2 = uqbar.graphs.Node()
            >>> g.extend([n1, n2])
            >>> e = n1.attach(n2)
        """,
        """
        <literal_block xml:space="preserve">
            >>> print(format(g, "graphviz"))
        """,
        """
        <literal_block xml:space="preserve">
            >>> uqbar.graphs.Grapher(g)()
        """,
    ]
    actual = [normalize(block.pformat()) for block in blocks]
    expected = [normalize(text) for text in expected]
    assert actual == expected


def test_interpret_code_blocks_01():
    document = uqbar.book.sphinx.parse_rst(source_a)
    blocks = uqbar.book.sphinx.collect_literal_blocks(document)
    node_mapping = uqbar.book.sphinx.interpret_code_blocks(blocks)
    assert list(node_mapping.values()) == [
        [ConsoleInput(string=">>> string = 'Hello, world!'\n")],
        [
            ConsoleInput(
                string='>>> for i in range(3):\n...     string += " {}".format(i)\n... \n'
            )
        ],
        [
            ConsoleInput(string=">>> print(string)\n"),
            ConsoleOutput(string="Hello, world! 0 1 2\n"),
        ],
    ]


def test_interpret_code_blocks_02():
    def logger_func(message):
        messages.append(message)

    error_message = (
        "Traceback (most recent call last):\n" '  File "<stdin>", line 1, in <module>\n'
    )
    if sys.version_info < (3, 7):
        error_message += "TypeError: must be str, not int\n"
    else:
        error_message += 'TypeError: can only concatenate str (not "int") to str\n'

    messages = []
    source = normalize(
        """
        This will interpret happily.

        ::

            >>> "1" + 2
            Traceback (most recent call last):
              File "<stdin>", line 1, in <module>
            {error_message}
        """.format(
            error_message=error_message
        )
    )
    document = uqbar.book.sphinx.parse_rst(source)
    blocks = uqbar.book.sphinx.collect_literal_blocks(document)
    # This has a traceback, so it passes.
    uqbar.book.sphinx.interpret_code_blocks(blocks, logger_func=logger_func)
    assert messages == [error_message]
    messages[:] = []
    source = normalize(
        """
        This will not interpret happily.

        ::

            >>> for i in range(1, 4):
            ...     i / 0
            ...
            "This is fine"
        """
    )
    document = uqbar.book.sphinx.parse_rst(source)
    blocks = uqbar.book.sphinx.collect_literal_blocks(document)
    with pytest.raises(ConsoleError):
        # This does not have a traceback, so it fails.
        uqbar.book.sphinx.interpret_code_blocks(blocks, logger_func=logger_func)
    assert messages == [
        (
            "Traceback (most recent call last):\n"
            '  File "<stdin>", line 2, in <module>\n'
            "ZeroDivisionError: division by zero\n"
        )
    ]
    messages[:] = []
    # This passes because we force it to.
    uqbar.book.sphinx.interpret_code_blocks(
        blocks, allow_exceptions=True, logger_func=logger_func
    )
    assert messages == [
        (
            "Traceback (most recent call last):\n"
            '  File "<stdin>", line 2, in <module>\n'
            "ZeroDivisionError: division by zero\n"
        )
    ]


def test_rebuild_document_01():
    document = uqbar.book.sphinx.parse_rst(source_a)
    blocks = uqbar.book.sphinx.collect_literal_blocks(document)
    node_mapping = uqbar.book.sphinx.interpret_code_blocks(blocks)
    uqbar.book.sphinx.rebuild_document(document, node_mapping)
    assert normalize(document.pformat()) == normalize(
        """
        <document source="test">
            <literal_block xml:space="preserve">
                >>> string = 'Hello, world!'
            <literal_block xml:space="preserve">
                >>> for i in range(3):
                ...     string += " {}".format(i)
                ...
            <literal_block xml:space="preserve">
                >>> print(string)
                Hello, world! 0 1 2
        """
    )


def test_rebuild_document_02():
    document = uqbar.book.sphinx.parse_rst(source_b)
    blocks = uqbar.book.sphinx.collect_literal_blocks(document)
    extensions = [GraphExtension]
    node_mapping = uqbar.book.sphinx.interpret_code_blocks(
        blocks, extensions=extensions
    )
    uqbar.book.sphinx.rebuild_document(document, node_mapping)
    assert normalize(document.pformat()) == normalize(
        """
        <document source="test">
            <literal_block xml:space="preserve">
                >>> import uqbar.graphs
                >>> g = uqbar.graphs.Graph()
                >>> n1 = uqbar.graphs.Node()
                >>> n2 = uqbar.graphs.Node()
                >>> g.extend([n1, n2])
                >>> e = n1.attach(n2)
            <literal_block xml:space="preserve">
                >>> print(format(g, "graphviz"))
                digraph G {
                    node_0;
                    node_1;
                    node_0 -> node_1;
                }
            <literal_block xml:space="preserve">
                >>> uqbar.graphs.Grapher(g)()
            <graphviz_block layout="dot" xml:space="preserve">
                digraph G {
                    node_0;
                    node_1;
                    node_0 -> node_1;
                }
        """
    )


def test_rebuild_document_03():
    document = uqbar.book.sphinx.parse_rst(source_c)
    blocks = uqbar.book.sphinx.collect_literal_blocks(document)
    extensions = [GraphExtension]
    node_mapping = uqbar.book.sphinx.interpret_code_blocks(
        blocks, extensions=extensions
    )
    uqbar.book.sphinx.rebuild_document(document, node_mapping)
    assert normalize(document.pformat()) == normalize(
        """
        <document source="test">
            <literal_block xml:space="preserve">
                >>> import uqbar.graphs
                >>> g = uqbar.graphs.Graph()
                >>> n1 = uqbar.graphs.Node()
                >>> g.append(n1)
                >>> for i in range(3):
                ...     n2 = uqbar.graphs.Node()
                ...     g.append(n2)
                ...     e = n1.attach(n2)
                ...     uqbar.graphs.Grapher(g)()
                ...     print(i)
                ...     n1 = n2
                ...
            <graphviz_block layout="dot" xml:space="preserve">
                digraph G {
                    node_0;
                    node_1;
                    node_0 -> node_1;
                }
            <literal_block xml:space="preserve">
                0
            <graphviz_block layout="dot" xml:space="preserve">
                digraph G {
                    node_0;
                    node_1;
                    node_2;
                    node_0 -> node_1;
                    node_1 -> node_2;
                }
            <literal_block xml:space="preserve">
                1
            <graphviz_block layout="dot" xml:space="preserve">
                digraph G {
                    node_0;
                    node_1;
                    node_2;
                    node_3;
                    node_0 -> node_1;
                    node_1 -> node_2;
                    node_2 -> node_3;
                }
            <literal_block xml:space="preserve">
                2
        """
    )


def test_rebuild_document_04():
    document = uqbar.book.sphinx.parse_rst(source_d)
    blocks = uqbar.book.sphinx.collect_literal_blocks(document)
    extensions = [GraphExtension]
    node_mapping = uqbar.book.sphinx.interpret_code_blocks(
        blocks, extensions=extensions
    )
    uqbar.book.sphinx.rebuild_document(document, node_mapping)
    assert normalize(document.pformat()) == normalize(
        """
        <document source="test">
            <graphviz_block layout="dot" xml:space="preserve">
                digraph G {
                    node_0;
                    node_1;
                    node_0 -> node_1;
                }
        """
    )



from typing import Dict, Iterable, List, Mapping, Optional, Set, Tuple, Union  # noqa

import uqbar.graphs

from ..containers import UniqueTreeList, UniqueTreeNode
from .attrs import Attributes


class Attachable(UniqueTreeNode):
    ### CLASS VARIABLES ###

    __documentation_section__ = "Mixins"

    ### INITIALIZER ###

    def __init__(self) -> None:
        self._edges: Set["Edge"] = set()

    ### PRIVATE METHODS ###

    def _get_canonical_name(self) -> str:
        node = self._get_node()
        node_name = ""
        if node is not None:
            node_name = node._get_canonical_name()
        port_name = self._get_port_name()
        return "{}:{}".format(node_name, port_name)

    def _get_node(self) -> Optional["Node"]:
        parent = self.parent
        while parent is not None:
            if isinstance(parent, Node):
                return parent
            parent = parent.parent
        return None

    def _get_port_name(self) -> str:
        graph_order = self.graph_order
        node = self._get_node()
        if node is not None:
            node_graph_order = node.graph_order
            graph_order = graph_order[len(node_graph_order) :]
        else:
            graph_order = (0,)
        return "f_" + "_".join(str(x) for x in graph_order)

    ### PUBLIC METHODS ###

    def attach(
        self,
        target,
        is_directed=True,
        head_port_position=None,
        tail_port_position=None,
        **attributes,
    ) -> "Edge":
        edge = Edge(
            attributes=attributes,
            head_port_position=head_port_position,
            is_directed=is_directed,
            tail_port_position=tail_port_position,
        )
        edge.attach(self, target)
        return edge

    ### PUBLIC PROPERTIES ###

    @property
    def edges(self) -> Set["Edge"]:
        return set(self._edges)


class Graph(UniqueTreeList):
    """
    A Graphviz graph, subgraph or cluster.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Core Classes"

    ### INITIALIZER ###

    def __init__(
        self,
        children=None,
        *,
        attributes: Optional[Union[Mapping[str, object], Attributes]] = None,
        edge_attributes: Optional[Union[Mapping[str, object], Attributes]] = None,
        is_cluster: bool = False,
        is_digraph: bool = True,
        name: Optional[str] = None,
        node_attributes: Optional[Union[Mapping[str, object], Attributes]] = None,
    ) -> None:
        UniqueTreeList.__init__(self, name=name, children=children)
        self._attributes = Attributes(
            "cluster" if is_cluster else "graph", **(attributes or {})
        )
        self._edge_attributes = Attributes("edge", **(edge_attributes or {}))
        self._node_attributes = Attributes("node", **(node_attributes or {}))
        self._is_cluster = bool(is_cluster)
        self._is_digraph = bool(is_digraph)

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: Optional[str] = None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == "graphviz":
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        def recurse(graph):
            indent = "    "
            result = []
            if not graph.parent:
                name = graph.name or "G"
                if graph.is_digraph:
                    string = "digraph {} {{".format(Attributes._format_value(name))
                else:
                    string = "graph {} {{".format(Attributes._format_value(name))
            else:
                if graph.name is not None:
                    name = graph.name
                    if graph.is_cluster:
                        name = "cluster_{}".format(name)
                else:
                    name = graph._get_canonical_name()
                string = "subgraph {} {{".format(Attributes._format_value(name))
            result.append(string)
            if graph.attributes:
                attributes = "graph {}".format(
                    format(graph.attributes, "graphviz")
                ).split("\n")
                result.extend(indent + line for line in attributes)
            if graph.node_attributes:
                attributes = "node {}".format(
                    format(graph.node_attributes, "graphviz")
                ).split("\n")
                result.extend(indent + line for line in attributes)
            if graph.edge_attributes:
                attributes = "edge {}".format(
                    format(graph.edge_attributes, "graphviz")
                ).split("\n")
                result.extend(indent + line for line in attributes)
            for child in graph:
                if isinstance(child, type(self)):
                    lines = (indent + line for line in recurse(child))
                else:
                    lines = (
                        indent + line for line in format(child, "graphviz").split("\n")
                    )
                result.extend(lines)
            if graph in edge_parents:
                for edge in edge_parents[graph]:
                    lines = format(edge, "graphviz").split("\n")
                    result.extend(indent + line for line in lines)
            result.append("}")
            return result

        all_edges: Set[Edge] = set()
        for child in self.depth_first():
            for edge in getattr(child, "edges", ()):
                if edge.tail.root is not edge.head.root:
                    continue
                all_edges.add(edge)
        edge_parents: Dict[Graph, List[Edge]] = {}
        for edge in sorted(
            all_edges, key=lambda edge: (edge.tail_graph_order, edge.head_graph_order)
        ):
            highest_parent = edge._get_highest_parent()
            edge_parents.setdefault(highest_parent, []).append(edge)

        return "\n".join(recurse(self))

    ### PRIVATE METHODS ###

    def _get_canonical_name(self) -> str:
        name_prefix = "graph"
        if self.is_cluster:
            name_prefix = "cluster"
        if self.name is not None:
            name = self.name
            root = self.root
            if root:
                instances = root[self.name]
                if not isinstance(instances, type(self)):
                    name = "{}_{}".format(name, instances.index(self))
            suffix = name
        elif self.graph_order:
            suffix = "_".join(str(x) for x in self.graph_order)
        else:
            suffix = "0"
        return "{}_{}".format(name_prefix, suffix)

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self) -> Tuple[type, ...]:
        return (Graph, Node)

    ### PUBLIC PROPERTIES ###

    @property
    def attributes(self) -> Attributes:
        return self._attributes

    @property
    def edge_attributes(self) -> Attributes:
        return self._edge_attributes

    @property
    def is_cluster(self) -> bool:
        return self._is_cluster

    @property
    def is_digraph(self) -> bool:
        return self._is_digraph

    @property
    def node_attributes(self) -> Attributes:
        return self._node_attributes


class Node(UniqueTreeList):
    """
    A Graphviz node.

    ::

        >>> import uqbar.graphs
        >>> node = uqbar.graphs.Node()
        >>> print(format(node, 'graphviz'))
        node_0;

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Core Classes"

    ### INITIALIZER ###

    def __init__(
        self,
        children: Optional[
            Iterable[Union["uqbar.graphs.RecordField", "uqbar.graphs.RecordGroup"]]
        ] = None,
        *,
        attributes: Optional[Union[Mapping[str, object], Attributes]] = None,
        name: Optional[str] = None,
    ) -> None:
        UniqueTreeList.__init__(self, name=name, children=children)
        self._attributes = Attributes("node", **(attributes or {}))
        self._edges: Set[Edge] = set()

    ### SPECIAL METHODS ###

    def __format__(self, format_spec=None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == "graphviz":
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        from .html import Table  # avoid circular imports

        node_definition = Attributes._format_value(self._get_canonical_name())
        result = [node_definition]
        attributes = self.attributes.copy()
        if len(self):
            if isinstance(self[0], Table):
                label = "<\n{}>".format(format(self[0], "graphviz"))
            else:
                label = " | ".join(format(_, "graphviz") for _ in self)
            attributes["label"] = label
        if len(attributes):
            attributes = format(attributes, "graphviz").split("\n")
            result[0] = "{} {}".format(result[0], attributes[0])
            result.extend(attributes[1:])
        else:
            result[-1] += ";"
        return "\n".join(result)

    ### PRIVATE METHODS ###

    def _get_canonical_name(self) -> str:
        prefix = "node"
        if self.name is not None:
            root = self.root
            if root:
                instances = root[self.name]
                if instances is not self:
                    return "{}_{}".format(self.name, instances.index(self))
            return self.name
        elif self.graph_order:
            suffix = "_".join(str(x) for x in self.graph_order)
        else:
            suffix = "0"
        return "{}_{}".format(prefix, suffix)

    ### PUBLIC METHODS ###

    def attach(
        self,
        node,
        is_directed=True,
        head_port_position=None,
        tail_port_position=None,
        **attributes,
    ) -> "Edge":
        edge = Edge(
            attributes=attributes,
            head_port_position=head_port_position,
            is_directed=is_directed,
            tail_port_position=tail_port_position,
        )
        edge.attach(self, node)
        return edge

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self) -> Tuple[type, ...]:
        from .html import Table  # avoid circular imports
        from .records import RecordField, RecordGroup  # avoid circular imports

        return (RecordField, RecordGroup, Table)

    ### PUBLIC PROPERTIES ###

    @property
    def attributes(self) -> Attributes:
        return self._attributes

    @property
    def edges(self) -> Set["Edge"]:
        return set(self._edges)


class Edge(object):
    """
    A Graphviz edge.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Core Classes"

    ### INITIALIZER ###

    def __init__(
        self,
        attributes: Optional[Union[Mapping[str, object], Attributes]] = None,
        is_directed: bool = True,
        head_port_position: Optional[str] = None,
        tail_port_position: Optional[str] = None,
    ) -> None:
        self._attributes = Attributes("edge", **(attributes or {}))
        self._head: Optional[Union[Node, Attachable]] = None
        self._head_port_position = head_port_position
        self._is_directed = bool(is_directed)
        self._tail: Optional[Union[Node, Attachable]] = None
        self._tail_port_position = tail_port_position

    ### SPECIAL METHODS ###

    def __format__(self, format_spec=None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == "graphviz":
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        connection = "->"
        if not self.is_directed:
            connection = "--"

        tail_parts = []
        tail_node = None
        tail_port_name = None
        if isinstance(self.tail, Node):
            tail_node = self.tail
        elif isinstance(self.tail, Attachable):
            tail_node = self.tail._get_node()
            tail_port_name = self.tail._get_port_name()
        if tail_node is None:
            raise ValueError
        tail_parts.append(tail_node._get_canonical_name())
        if tail_port_name:
            tail_parts.append(tail_port_name)
        if self.tail_port_position:
            tail_parts.append(self.tail_port_position)
        tail_name = ":".join(Attributes._format_value(part) for part in tail_parts)

        head_parts = []
        head_node = None
        head_port_name = None
        if isinstance(self.head, Node):
            head_node = self.head
        elif isinstance(self.head, Attachable):
            head_node = self.head._get_node()
            head_port_name = self.head._get_port_name()
        if head_node is None:
            raise ValueError
        head_parts.append(head_node._get_canonical_name())
        if head_port_name:
            head_parts.append(head_port_name)
        if self.head_port_position:
            head_parts.append(self.head_port_position)
        head_name = ":".join(Attributes._format_value(part) for part in head_parts)

        edge_definition = "{} {} {}".format(tail_name, connection, head_name)
        result = [edge_definition]
        if len(self.attributes):
            attributes = format(self.attributes, "graphviz").split("\n")
            result[0] = "{} {}".format(result[0], attributes[0])
            result.extend(attributes[1:])
        else:
            result[-1] += ";"
        return "\n".join(result)

    ### PRIVATE METHODS ###

    def _get_highest_parent(self) -> Graph:
        if self.tail is None:
            raise ValueError(self.tail)
        elif self.head is None:
            raise ValueError(self.head)
        highest_parent: Optional[Graph] = None
        tail_parentage = list(self.tail.parentage[1:])
        head_parentage = list(self.head.parentage[1:])
        while (
            len(tail_parentage)
            and len(head_parentage)
            and tail_parentage[-1] is head_parentage[-1]
        ):
            highest_parent = tail_parentage[-1]
            tail_parentage.pop()
            head_parentage.pop()
        if highest_parent is None:
            message = "highest parent can not be none."
            raise Exception(message)
        return highest_parent

    ### PUBLIC METHODS ###

    def attach(
        self, tail: Union[Node, Attachable], head: Union[Node, Attachable]
    ) -> "Edge":
        prototype = (Node, Attachable)
        assert isinstance(tail, prototype)
        assert isinstance(head, prototype)
        self.detach()
        tail._edges.add(self)
        head._edges.add(self)
        self._tail = tail
        self._head = head
        return self

    def detach(self) -> "Edge":
        if self.tail is not None:
            self.tail._edges.remove(self)
            self._tail = None
        if self.head is not None:
            self.head._edges.remove(self)
            self._head = None
        return self

    ### PUBLIC PROPERTIES ###

    @property
    def attributes(self) -> Attributes:
        return self._attributes

    @property
    def head(self) -> Optional[Union[Node, Attachable]]:
        return self._head

    @property
    def head_graph_order(self) -> Tuple[int, ...]:
        if self.head is None:
            return ()
        return self.head.graph_order

    @property
    def head_port_position(self) -> Optional[str]:
        return self._head_port_position

    @property
    def is_directed(self) -> bool:
        return self._is_directed

    @property
    def tail(self) -> Optional[Union[Node, Attachable]]:
        return self._tail

    @property
    def tail_graph_order(self) -> Tuple[int, ...]:
        if self.tail is None:
            return ()
        return self.tail.graph_order

    @property
    def tail_port_position(self) -> Optional[str]:
        return self._tail_port_position



from uqbar.graphs import Graph, LineBreak, Node, Table, TableCell, TableRow, Text
from uqbar.strings import normalize


def test_graphs_Table():
    """
    digraph structs {

        node [shape=plaintext]


        struct1:f1 -> struct2:f0;
        struct1:f2 -> struct3:here;
    }
    """

    graph = Graph()
    graph.node_attributes["shape"] = "plaintext"

    struct_1_table_row_1_cell_1 = TableCell("left")
    struct_1_table_row_1_cell_2 = TableCell("mid dle")
    struct_1_table_row_1_cell_3 = TableCell("right")
    struct_1_table_row_1 = TableRow(
        [
            struct_1_table_row_1_cell_1,
            struct_1_table_row_1_cell_2,
            struct_1_table_row_1_cell_3,
        ]
    )
    struct_1_table = Table(
        [struct_1_table_row_1],
        attributes={"border": 0, "cellborder": 1, "cellspacing": 0},
    )
    struct_1 = Node([struct_1_table])

    struct_2_table_row_1_cell_1 = TableCell("one")
    struct_2_table_row_1_cell_2 = TableCell("two")
    struct_2_table_row_1 = TableRow(
        [struct_2_table_row_1_cell_1, struct_2_table_row_1_cell_2]
    )
    struct_2_table = Table(
        [struct_2_table_row_1],
        attributes={"border": 0, "cellborder": 1, "cellspacing": 0},
    )
    struct_2 = Node([struct_2_table])

    struct_3_table_row_1_cell_1 = TableCell(
        [Text("hello"), LineBreak(), Text("world")], attributes={"rowspan": 3}
    )
    struct_3_table_row_1_cell_2 = TableCell("b", attributes={"colspan": 3})
    struct_3_table_row_1_cell_3 = TableCell("g", attributes={"rowspan": 3})
    struct_3_table_row_1_cell_4 = TableCell("h", attributes={"rowspan": 3})
    struct_3_table_row_1 = TableRow(
        [
            struct_3_table_row_1_cell_1,
            struct_3_table_row_1_cell_2,
            struct_3_table_row_1_cell_3,
            struct_3_table_row_1_cell_4,
        ]
    )

    struct_3_table_row_2_cell_1 = TableCell("c")
    struct_3_table_row_2_cell_2 = TableCell("d")
    struct_3_table_row_2_cell_3 = TableCell("e")
    struct_3_table_row_2 = TableRow(
        [
            struct_3_table_row_2_cell_1,
            struct_3_table_row_2_cell_2,
            struct_3_table_row_2_cell_3,
        ]
    )

    struct_3_table_row_3_cell_1 = TableCell("f", attributes={"colspan": 3})

    struct_3_table_row_3 = TableRow([struct_3_table_row_3_cell_1])

    struct_3_table = Table(
        [struct_3_table_row_1, struct_3_table_row_2, struct_3_table_row_3],
        attributes={"border": 0, "cellborder": 1, "cellspacing": 0, "cellpadding": 4},
    )
    struct_3 = Node([struct_3_table])

    graph.extend([struct_1, struct_2, struct_3])

    struct_1_table_row_1_cell_2.attach(struct_2_table_row_1_cell_1)
    struct_1_table_row_1_cell_3.attach(struct_3_table_row_2_cell_2)

    graphviz_format = format(graph, "graphviz")

    assert graphviz_format == normalize(
        """
        digraph G {
            node [shape=plaintext];
            node_0 [label=<
                <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0">
                    <TR>
                        <TD>left</TD>
                        <TD PORT="f_0_0_1">mid dle</TD>
                        <TD PORT="f_0_0_2">right</TD>
                    </TR>
                </TABLE>>];
            node_1 [label=<
                <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0">
                    <TR>
                        <TD PORT="f_0_0_0">one</TD>
                        <TD>two</TD>
                    </TR>
                </TABLE>>];
            node_2 [label=<
                <TABLE BORDER="0" CELLBORDER="1" CELLPADDING="4" CELLSPACING="0">
                    <TR>
                        <TD ROWSPAN="3">hello<BR/>world</TD>
                        <TD COLSPAN="3">b</TD>
                        <TD ROWSPAN="3">g</TD>
                        <TD ROWSPAN="3">h</TD>
                    </TR>
                    <TR>
                        <TD>c</TD>
                        <TD PORT="f_0_1_1">d</TD>
                        <TD>e</TD>
                    </TR>
                    <TR>
                        <TD COLSPAN="3">f</TD>
                    </TR>
                </TABLE>>];
            node_0:f_0_0_1 -> node_1:f_0_0_0;
            node_0:f_0_0_2 -> node_2:f_0_1_1;
        }
        """
    )



import pathlib
import shutil
import sys

import pytest
import sphinx.errors

from uqbar.strings import ansi_escape, normalize

index_content = normalize(
    """
    Fake Docs
    *********

       >>> print("hello, world!");
       hello, world!

       >>> print("foo bar baz!");
       foo bar baz!

       >>> print("i'm on my own line too!");
       i'm on my own line too!

       >>> import uqbar.graphs

       >>> g = uqbar.graphs.Graph()
       >>> n1 = uqbar.graphs.Node()
       >>> g.append(n1)

       >>> for i in range(3):
       ...     n2 = uqbar.graphs.Node()
       ...     g.append(n2)
       ...     e = n1.attach(n2)
       ...     uqbar.graphs.Grapher(g)()
       ...     print(i)
       ...     n1 = n2
       ...

       digraph G {
           node_0;
           node_1;
           node_0 -> node_1;
       }

       0

       digraph G {
           node_0;
           node_1;
           node_2;
           node_0 -> node_1;
           node_1 -> node_2;
       }

       1

       digraph G {
           node_0;
           node_1;
           node_2;
           node_3;
           node_0 -> node_1;
           node_1 -> node_2;
           node_2 -> node_3;
       }

       2

       >>> print(format(g, "graphviz"))
       digraph G {
           node_0;
           node_1;
           node_2;
           node_3;
           node_0 -> node_1;
           node_1 -> node_2;
           node_2 -> node_3;
       }

    * API

      * "just_a_function()"

      * "GrandParent"

        * "GrandParent.one()"

        * "GrandParent.three()"

        * "GrandParent.two()"

      * "Parent"

        * "Parent.one()"

        * "Parent.three()"

        * "Parent.two()"

      * "Uncle"

        * "Uncle.one()"

        * "Uncle.three()"

        * "Uncle.two()"

      * "Child"

        * "Child.one()"

        * "Child.three()"

        * "Child.two()"

      * "Outer"

        * "Outer.Inner"

          * "Outer.Inner.inner_method()"

        * "Outer.outer_method()"

    * Directives
    """
)

api_content = normalize(
    """
    API
    ***

    fake.just_a_function()

       Just a function.

          >>> print("I am just a function.")
          I am just a function.

    Text:

       >>> print("I am text.")
       I am text.

    class fake.GrandParent

       A grand parent.

          >>> print("I am a grand parent.")
          I am a grand parent.

       one()

          Grand parent: one.

             >>> print("Grand parent: one.")
             Grand parent: one.

       three()

          Grand parent: three.

             >>> print("Grand parent: three.")
             Grand parent: three.

             >>> import uqbar.graphs
             >>> g = uqbar.graphs.Graph()
             >>> n1 = uqbar.graphs.Node()
             >>> n2 = uqbar.graphs.Node()
             >>> g.extend([n1, n2])
             >>> e = n1.attach(n2)

             >>> print(format(g, "graphviz"))
             digraph G {
                 node_0;
                 node_1;
                 node_0 -> node_1;
             }

             >>> uqbar.graphs.Grapher(g)()

             digraph G {
                 node_0;
                 node_1;
                 node_0 -> node_1;
             }

          And that, my friends, is a graph.

       two()

          Grand parent: two.

             >>> print("Grand parent: two.")
             Grand parent: two.

    Text:

       >>> print("I am text.")
       I am text.

    class fake.Parent

       A parent.

          >>> print("I am a parent.")
          I am a parent.

       one()

          Parent: one.

             >>> print("Parent: one.")
             Parent: one.

       three()

          Grand parent: three.

             >>> print("Grand parent: three.")
             Grand parent: three.

             >>> import uqbar.graphs
             >>> g = uqbar.graphs.Graph()
             >>> n1 = uqbar.graphs.Node()
             >>> n2 = uqbar.graphs.Node()
             >>> g.extend([n1, n2])
             >>> e = n1.attach(n2)

             >>> print(format(g, "graphviz"))
             digraph G {
                 node_0;
                 node_1;
                 node_0 -> node_1;
             }

             >>> uqbar.graphs.Grapher(g)()

             digraph G {
                 node_0;
                 node_1;
                 node_0 -> node_1;
             }

          And that, my friends, is a graph.

       two()

          Grand parent: two.

             >>> print("Grand parent: two.")
             Grand parent: two.

    Text:

       >>> print("I am text.")
       I am text.

    class fake.Uncle

       one()

          Grand parent: one.

             >>> print("Grand parent: one.")
             Grand parent: one.

       three()

          Grand parent: three.

             >>> print("Grand parent: three.")
             Grand parent: three.

             >>> import uqbar.graphs
             >>> g = uqbar.graphs.Graph()
             >>> n1 = uqbar.graphs.Node()
             >>> n2 = uqbar.graphs.Node()
             >>> g.extend([n1, n2])
             >>> e = n1.attach(n2)

             >>> print(format(g, "graphviz"))
             digraph G {
                 node_0;
                 node_1;
                 node_0 -> node_1;
             }

             >>> uqbar.graphs.Grapher(g)()

             digraph G {
                 node_0;
                 node_1;
                 node_0 -> node_1;
             }

          And that, my friends, is a graph.

       two()

          Grand parent: two.

             >>> print("Grand parent: two.")
             Grand parent: two.

    Text:

       >>> print("I am text.")
       I am text.

    class fake.Child

       A child.

          >>> print("I am a child.")
          I am a child.

       one()

          Parent: one.

             >>> print("Parent: one.")
             Parent: one.

       three()

          Grand parent: three.

             >>> print("Grand parent: three.")
             Grand parent: three.

             >>> import uqbar.graphs
             >>> g = uqbar.graphs.Graph()
             >>> n1 = uqbar.graphs.Node()
             >>> n2 = uqbar.graphs.Node()
             >>> g.extend([n1, n2])
             >>> e = n1.attach(n2)

             >>> print(format(g, "graphviz"))
             digraph G {
                 node_0;
                 node_1;
                 node_0 -> node_1;
             }

             >>> uqbar.graphs.Grapher(g)()

             digraph G {
                 node_0;
                 node_1;
                 node_0 -> node_1;
             }

          And that, my friends, is a graph.

       two()

          Child: two.

             >>> print("Child: two.")
             Child: two.

    Text:

       >>> print("I am text.")
       I am text.

    class fake.Outer

       An outer class.

          >>> print("I am an outer class.")
          I am an outer class.

       class Inner

          An inner class.

             >>> print("I am an inner class.")
             I am an inner class.

          inner_method()

             Inner: method.

                >>> print("Inner: method.")
                Inner: method.

       outer_method()

          Outer: method.

             >>> print("Outer: method.")
             Outer: method.
    """
)

directives_content = normalize(
    """
    Directives
    **********

       >>> value = 1

       >>> value += 1
       >>> value
       2

       class ChildClass(PublicClass):
           def inheritable_method(self):
               pass

           def new_method(self):
               pass

       >>> value += 1
       >>> value
       4

       >>> ChildClass.__name__
       'ChildClass'

       >>> 1 / 0
       Traceback (most recent call last):
         File "<stdin>", line 1, in <module>
       ZeroDivisionError: division by zero
    """
)


@pytest.fixture(autouse=True)
def test_path():
    path = pathlib.Path(__file__).parent / "roots" / "test-uqbar-sphinx-book"
    assert path.exists()
    sys.path.insert(0, str(path))
    yield
    sys.path.remove(str(path))


@pytest.fixture
def rm_dirs(app):
    path = pathlib.Path(app.doctreedir)
    shutil.rmtree(path)
    yield


@pytest.mark.sphinx(
    "html", testroot="uqbar-sphinx-book", confoverrides={"uqbar_book_use_cache": True}
)
def test_sphinx_book_html_cached(app, status, warning, rm_dirs):
    app.build()
    assert not warning.getvalue().strip()
    image_path = pathlib.Path(app.outdir) / "_images"
    dot_paths = [path for path in image_path.iterdir() if path.suffix == ".dot"]
    dot_sources = set(normalize(path.read_text()) for path in dot_paths)
    assert dot_sources == set(
        [
            normalize(
                """
                digraph G {
                    node_0;
                    node_1;
                    node_0 -> node_1;
                }
                """
            ),
            normalize(
                """
                digraph G {
                    node_0;
                    node_1;
                    node_2;
                    node_0 -> node_1;
                    node_1 -> node_2;
                }
                """
            ),
            normalize(
                """
                digraph G {
                    node_0;
                    node_1;
                    node_2;
                    node_3;
                    node_0 -> node_1;
                    node_1 -> node_2;
                    node_2 -> node_3;
                }
                """
            ),
        ]
    )
    for dot_path in dot_paths:
        assert dot_path.with_suffix(".svg").exists()


@pytest.mark.sphinx(
    "latex", testroot="uqbar-sphinx-book", confoverrides={"uqbar_book_use_cache": True}
)
def test_sphinx_book_latex_cached(app, status, warning, rm_dirs):
    app.build()
    assert not warning.getvalue().strip()


@pytest.mark.sphinx(
    "text", testroot="uqbar-sphinx-book", confoverrides={"uqbar_book_use_cache": True}
)
def test_sphinx_book_text_cached(app, status, warning, rm_dirs):
    app.build()
    assert not warning.getvalue().strip()
    assert app.config["uqbar_book_use_cache"]
    assert list(
        app.connection.execute("SELECT path, hits FROM cache ORDER BY path")
    ) == [
        ("fake.Child", 0),
        ("fake.Child.two", 0),
        ("fake.GrandParent", 0),
        ("fake.GrandParent.one", 1),
        ("fake.GrandParent.three", 3),
        ("fake.GrandParent.two", 2),
        ("fake.Outer", 0),
        ("fake.Outer.Inner", 0),
        ("fake.Outer.Inner.inner_method", 0),
        ("fake.Outer.outer_method", 0),
        ("fake.Parent", 0),
        ("fake.Parent.one", 1),
        ("fake.just_a_function", 0),
    ]
    assert not warning.getvalue().strip()
    for filename, expected_content in [
        ("api.txt", api_content),
        ("directives.txt", directives_content),
        ("index.txt", index_content),
    ]:
        path = pathlib.Path(app.srcdir) / "_build" / "text" / filename
        actual_content = normalize(path.read_text())
        assert actual_content == expected_content


@pytest.mark.sphinx(
    "text", testroot="uqbar-sphinx-book", confoverrides={"uqbar_book_use_cache": False}
)
def test_sphinx_book_text_uncached(app, status, warning, rm_dirs):
    app.build()
    assert not warning.getvalue().strip()
    assert not app.config["uqbar_book_use_cache"]
    for filename, expected_content in [
        ("api.txt", api_content),
        ("directives.txt", directives_content),
        ("index.txt", index_content),
    ]:
        path = pathlib.Path(app.srcdir) / "_build" / "text" / filename
        actual_content = normalize(path.read_text())
        assert actual_content == expected_content


@pytest.mark.sphinx("text", testroot="uqbar-sphinx-book-broken")
def test_sphinx_book_text_broken_strict(app, status, warning, rm_dirs):
    """
    Build halts.
    """
    with pytest.raises(sphinx.errors.ExtensionError):
        app.build()
    assert normalize(ansi_escape(status.getvalue())) == normalize(
        """
        Running Sphinx v{sphinx_version}
        [uqbar-book] initializing cache db
        building [mo]: targets for 0 po files that are out of date{writing}
        building [text]: targets for 1 source files that are out of date
        updating environment: [new config] 1 added, 0 changed, 0 removed
        reading sources... [100%] index
        """.format(
            sphinx_version=sphinx.__version__,
            writing="\n\t\twriting output..." if sphinx.version_info > (6,) else "",
        )
    )
    assert normalize(ansi_escape(warning.getvalue())) == normalize(
        f"""
        {app.srcdir / "index.rst"}:15: WARNING:
            <literal_block xml:space="preserve">>>> print(this_name_does_not_exist)</literal_block>
            Traceback (most recent call last):
              File "<stdin>", line 1, in <module>
            NameError: name 'this_name_does_not_exist' is not defined
        """
    )


@pytest.mark.sphinx(
    "text",
    testroot="uqbar-sphinx-book",
    confoverrides={"uqbar_book_console_setup": ["1 / 0"]},
)
def test_sphinx_book_text_broken_setup(make_app, app_params):
    """
    Fail build immediately on console setup error.
    """
    args, kwargs = app_params
    with pytest.raises(sphinx.errors.ExtensionError):
        make_app(*args, **kwargs)


@pytest.mark.sphinx(
    "text",
    testroot="uqbar-sphinx-book",
    confoverrides={"uqbar_book_console_teardown": ["1 / 0"]},
)
def test_sphinx_book_text_broken_teardown(make_app, app_params):
    """
    Fail build immediately on console teardown error.
    """
    args, kwargs = app_params
    with pytest.raises(sphinx.errors.ExtensionError):
        make_app(*args, **kwargs)



import pathlib
import sys

import pytest

import uqbar.apis
from uqbar.strings import normalize


@pytest.fixture
def test_path():
    test_path = pathlib.Path(__file__).parent
    if str(test_path) not in sys.path:
        sys.path.insert(0, str(test_path))


def test_str_01(test_path):
    documenter = uqbar.apis.ClassDocumenter("fake_package.module.PublicClass")
    assert normalize(str(documenter)) == normalize(
        """
        .. autoclass:: PublicClass
           :members:
           :undoc-members:
        """
    )


def test_str_02(test_path):
    documenter = uqbar.apis.ClassDocumenter("fake_package.module.ChildClass")
    assert normalize(str(documenter)) == normalize(
        """
        .. autoclass:: ChildClass
           :members:
           :undoc-members:
        """
    )


def test_str_03(test_path):
    documenter = uqbar.apis.ClassDocumenter("fake_package.module._PrivateClass")
    assert normalize(str(documenter)) == normalize(
        """
        .. autoclass:: _PrivateClass
           :members:
           :undoc-members:
        """
    )



import unittest

import uqbar.graphs
from uqbar.strings import normalize


class TestCase(unittest.TestCase):
    def test_mode(self):
        uqbar.graphs.Attributes(mode="cluster")
        uqbar.graphs.Attributes(mode="edge")
        uqbar.graphs.Attributes(mode="graph")
        uqbar.graphs.Attributes(mode="node")

    def test___format___str_01(self):
        attributes = uqbar.graphs.Attributes(mode="node")
        assert format(attributes) == repr(attributes)

    def test___format___str_02(self):
        attributes = uqbar.graphs.Attributes(
            mode="node",
            color="blue",
            fontname="Times New Roman",
            fontsize=11.5,
            shape="oval",
        )
        assert format(attributes) == repr(attributes)

    def test___format___graphviz_01(self):
        attributes = uqbar.graphs.Attributes(mode="node")
        assert format(attributes, "graphviz") == ""

    def test___format___graphviz_02(self):
        attributes = uqbar.graphs.Attributes(
            mode="node",
            color="blue",
            fontname="Times New Roman",
            fontsize=11.5,
            shape="oval",
        )
        assert format(attributes, "graphviz") == normalize(
            """
            [color=blue,
                fontname="Times New Roman",
                fontsize=11.5,
                shape=oval];
            """
        )



import platform
import sys
from pathlib import Path

import uqbar.io


def test_find_executable():
    print(f"{sys.executable=}")
    file_name = "python"
    if platform.system() == "Windows":
        file_name = "python.exe"
    found = uqbar.io.find_executable(file_name)
    sys_executable = str(Path(sys.executable).resolve())
    print(f"{sys_executable=}")
    print(f"{found=}")
    assert sys_executable in found



"""
Uqbar Sphinx API generation extension.

Install by adding ``'uqbar.sphinx.api'`` to the ``extensions`` list in your
Sphinx configuration.

This extension provides the following configuration values which correspond to
the initialization arguments to the :py:class:`uqbar.apis.APIBuilder` class.

- ``uqbar_api_directory_name`` (default: ``api``)
- ``uqbar_api_document_empty_modules`` (default: ``False``)
- ``uqbar_api_document_private_members`` (default: ``False``)
- ``uqbar_api_document_private_modules`` (default: ``False``)
- ``uqbar_api_member_documenter_classes`` (default: ``None``)
- ``uqbar_api_module_documenter_class`` (default: ``None``)
- ``uqbar_api_root_documenter_class`` (default: ``None``)
- ``uqbar_api_source_paths`` (default: ``[]``)
- ``uqbar_api_title`` (default: ``API``)

reStructuredText source files will be generated for the modules given by
``uqbar_api_source_paths`` in the directory ``uqbar_api_directory_name``
relative to your Sphinx source directory.
"""

import importlib
import pathlib
import types
from typing import Any, Dict, List

from sphinx.util import logging
from sphinx.util.console import bold, darkgreen, darkred, purple  # type: ignore

import uqbar.apis

logger = logging.getLogger(__name__)


def logger_func(string):
    if string.startswith("preserved"):
        return
    elif string.startswith("rewrote"):
        string = purple(string)
    elif string.startswith("pruned"):
        string = darkred(string)
    elif string.startswith("wrote"):
        string = darkgreen(string)
    logger.info("{} {}".format(bold("[uqbar-api]"), string))


def on_builder_inited(app) -> None:
    """
    Hooks into Sphinx's ``builder-inited`` event.

    Builds out the ReST API source.
    """
    config = app.builder.config
    target_directory = (
        pathlib.Path(app.builder.env.srcdir) / config.uqbar_api_directory_name
    )
    initial_source_paths: List[str] = []
    source_paths = config.uqbar_api_source_paths
    for source_path in source_paths:
        if isinstance(source_path, types.ModuleType):
            if hasattr(source_path, "__path__") and source_path.__path__ is not None:
                initial_source_paths.extend(source_path.__path__)
            elif hasattr(source_path, "__file__") and source_path.__file__ is not None:
                initial_source_paths.extend(source_path.__file__)
            continue
        try:
            module = importlib.import_module(source_path)
            if hasattr(module, "__path__") and module.__path__ is not None:
                initial_source_paths.extend(module.__path__)
            elif hasattr(module, "__file__") and module.__file__ is not None:
                initial_source_paths.append(module.__file__)
        except ImportError:
            initial_source_paths.append(source_path)
    root_documenter_class = config.uqbar_api_root_documenter_class
    if isinstance(root_documenter_class, str):
        module_name, _, class_name = root_documenter_class.rpartition(".")
        module = importlib.import_module(module_name)
        root_documenter_class = getattr(module, class_name)
    module_documenter_class = config.uqbar_api_module_documenter_class
    if isinstance(module_documenter_class, str):
        module_name, _, class_name = module_documenter_class.rpartition(".")
        module = importlib.import_module(module_name)
        module_documenter_class = getattr(module, class_name)
    # Don't modify the list in Sphinx's config. Sphinx won't pickle class
    # references, and strips them from the saved config. That leads to Sphinx
    # believing that the config has changed on every run.
    member_documenter_classes = list(config.uqbar_api_member_documenter_classes or [])
    for i, member_documenter_class in enumerate(member_documenter_classes):
        if isinstance(member_documenter_class, str):
            module_name, _, class_name = member_documenter_class.rpartition(".")
            module = importlib.import_module(module_name)
            member_documenter_classes[i] = getattr(module, class_name)
    api_builder = uqbar.apis.APIBuilder(
        initial_source_paths=initial_source_paths,
        target_directory=target_directory,
        document_empty_modules=config.uqbar_api_document_empty_modules,
        document_private_members=config.uqbar_api_document_private_members,
        document_private_modules=config.uqbar_api_document_private_modules,
        member_documenter_classes=member_documenter_classes or None,
        module_documenter_class=module_documenter_class,
        omit_root=config.uqbar_api_omit_root,
        root_documenter_class=root_documenter_class,
        title=config.uqbar_api_title,
        logger_func=logger_func,
    )
    api_builder()


def setup(app) -> Dict[str, Any]:
    """
    Sets up Sphinx extension.
    """
    app.add_config_value("uqbar_api_directory_name", "api", "env")
    app.add_config_value("uqbar_api_document_empty_modules", False, "env")
    app.add_config_value("uqbar_api_document_private_members", False, "env")
    app.add_config_value("uqbar_api_document_private_modules", False, "env")
    app.add_config_value("uqbar_api_member_documenter_classes", None, "env")
    app.add_config_value("uqbar_api_module_documenter_class", None, "env")
    app.add_config_value("uqbar_api_omit_root", False, "env")
    app.add_config_value("uqbar_api_root_documenter_class", None, "env")
    app.add_config_value("uqbar_api_source_paths", None, "env")
    app.add_config_value("uqbar_api_title", "API", "html")
    app.connect("builder-inited", on_builder_inited)
    return {
        "version": uqbar.__version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import enum
import inspect
import textwrap
from typing import List, MutableMapping, Tuple, cast
from unittest import mock

from sphinx.ext.autosummary import extract_summary  # type: ignore
from sphinx.util.docutils import new_document  # type: ignore

from ..strings import normalize
from .documenters import (
    ClassDocumenter,
    MemberDocumenter,
    ModuleDocumenter,
    RootDocumenter,
)


class SummarizingClassDocumenter(ClassDocumenter):
    """
    A summarizing class documenter.

    Organizes class members by category, separated by category title and
    horizontal rule.

    Categories include:

    -  Special methods
    -  Methods
    -  Class and static methods
    -  Read/write properties
    -  Read-only properties

    ::

        >>> import uqbar.apis
        >>> path = 'uqbar.apis.summarizers.SummarizingClassDocumenter'
        >>> documenter = uqbar.apis.SummarizingClassDocumenter(path)
        >>> documentation = str(documenter)
        >>> print(documentation)
        .. autoclass:: SummarizingClassDocumenter
        <BLANKLINE>
           .. raw:: html
        <BLANKLINE>
              <hr/>
        <BLANKLINE>
           .. rubric:: Attributes Summary
              :class: class-header
        <BLANKLINE>
           .. autosummary::
              :nosignatures:
        <BLANKLINE>
              __str__
              ignored_special_methods
        <BLANKLINE>
           .. raw:: html
        <BLANKLINE>
              <hr/>
        <BLANKLINE>
           .. rubric:: Special methods
              :class: class-header
        <BLANKLINE>
           .. automethod:: SummarizingClassDocumenter.__str__
        <BLANKLINE>
           .. raw:: html
        <BLANKLINE>
              <hr/>
        <BLANKLINE>
           .. rubric:: Class & static methods
              :class: class-header
        <BLANKLINE>
           .. container:: inherited
        <BLANKLINE>
              .. automethod:: SummarizingClassDocumenter.validate_client
        <BLANKLINE>
           .. raw:: html
        <BLANKLINE>
              <hr/>
        <BLANKLINE>
           .. rubric:: Read-only properties
              :class: class-header
        <BLANKLINE>
           .. container:: inherited
        <BLANKLINE>
              .. autoattribute:: SummarizingClassDocumenter.client
        <BLANKLINE>
           .. container:: inherited
        <BLANKLINE>
              .. autoattribute:: SummarizingClassDocumenter.documentation_section
        <BLANKLINE>
           .. container:: inherited
        <BLANKLINE>
              .. autoattribute:: SummarizingClassDocumenter.package_path

    :param package_path: the module path and name of the member to document
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Documenters"

    ignored_special_methods: Tuple[str, ...] = (
        "__dict__",
        "__getattribute__",
        "__getnewargs__",
        "__getstate__",
        "__init__",
        "__reduce__",
        "__reduce_ex__",
        "__setstate__",
        "__sizeof__",
        "__subclasshook__",
        "fromkeys",
        "pipe_cloexec",
    )

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        name = getattr(self.client, "__name__")
        if issubclass(self.client, Exception):  # type: ignore
            return ".. autoexception:: {}".format(name)
        attributes = self._classify_class_attributes()
        (
            class_methods,
            data,
            methods,
            readonly_properties,
            readwrite_properties,
            special_methods,
            static_methods,
        ) = attributes
        result = [".. autoclass:: {}".format(name)]
        if issubclass(self.client, enum.Enum):  # type: ignore
            result.extend(["   :members:", "   :undoc-members:"])
        else:
            result.extend(self._build_member_autosummary(attributes))
        result.extend(
            self._build_attribute_section(
                special_methods, "automethod", "Special methods"
            )
        )
        result.extend(self._build_attribute_section(methods, "automethod", "Methods"))
        result.extend(
            self._build_attribute_section(
                sorted(class_methods + static_methods, key=lambda x: x.name),
                "automethod",
                "Class & static methods",
            )
        )
        result.extend(
            self._build_attribute_section(
                readwrite_properties, "autoattribute", "Read/write properties"
            )
        )
        result.extend(
            self._build_attribute_section(
                readonly_properties, "autoattribute", "Read-only properties"
            )
        )
        return "\n".join(result)

    ### PRIVATE METHODS ###

    def _build_attribute_section(
        self, attributes, directive: str, title: str
    ) -> List[str]:
        result: List[str] = []
        if not attributes:
            return result
        result.extend(
            [
                "",
                "   .. raw:: html",
                "",
                "      <hr/>",
                "",
                "   .. rubric:: {}".format(title),
                "      :class: class-header",
            ]
        )
        for attribute in attributes:
            result.append("")
            autodoc_directive = "   .. {}:: {}.{}".format(
                directive, getattr(self.client, "__name__"), attribute.name
            )
            if attribute.defining_class is self.client:
                result.append(autodoc_directive)
            else:
                result.append("   .. container:: inherited")
                result.append("")
                result.append("   {}".format(autodoc_directive))
        return result

    def _build_member_autosummary(self, attributes) -> List[str]:
        result: List[str] = []
        all_attributes: List[inspect.Attribute] = []
        for attribute_section in attributes:
            all_attributes.extend(
                attribute
                for attribute in attribute_section
                if attribute.defining_class is self.client
            )
        all_attributes.sort(key=lambda x: x.name)
        if not all_attributes:
            return result
        result.extend(
            [
                "",
                "   .. raw:: html",
                "",
                "      <hr/>",
                "",
                "   .. rubric:: {}".format("Attributes Summary"),
                "      :class: class-header",
                "",
                "   .. autosummary::",
                "      :nosignatures:",
                "",
            ]
        )
        for attribute in all_attributes:
            result.append("      {}".format(attribute.name))
        return result

    def _classify_class_attributes(self):
        class_methods = []
        data = []
        methods = []
        readonly_properties = []
        readwrite_properties = []
        special_methods = []
        static_methods = []
        attrs = inspect.classify_class_attrs(self.client)
        for attr in attrs:
            if attr.defining_class is object:
                continue
            elif (
                getattr(self.client, "__documentation_ignore_inherited__", None)
                and attr.defining_class is not self.client
            ):
                continue
            # Handle un-gettable attrs like Flask-SQLAlchemy's Model's `query`
            try:
                getattr(self.client, attr.name)
            except Exception:
                continue
            if attr.kind == "method":
                if attr.name not in self.ignored_special_methods:
                    if attr.name.startswith("__"):
                        special_methods.append(attr)
                    elif not attr.name.startswith("_"):
                        methods.append(attr)
            elif attr.kind == "class method":
                if attr.name not in self.ignored_special_methods:
                    if attr.name.startswith("__"):
                        special_methods.append(attr)
                    elif not attr.name.startswith("_"):
                        class_methods.append(attr)
            elif attr.kind == "static method":
                if attr.name not in self.ignored_special_methods:
                    if attr.name.startswith("__"):
                        special_methods.append(attr)
                    elif not attr.name.startswith("_"):
                        static_methods.append(attr)
            elif attr.kind == "property" and not attr.name.startswith("_"):
                if attr.object.fset is None:
                    readonly_properties.append(attr)
                else:
                    readwrite_properties.append(attr)
            elif (
                attr.kind == "data"
                and not attr.name.startswith("_")
                and attr.name not in getattr(self.client, "__slots__", ())
            ):
                data.append(attr)
        class_methods = tuple(sorted(class_methods))
        data = tuple(sorted(data))
        methods = tuple(sorted(methods))
        readonly_properties = tuple(sorted(readonly_properties))
        readwrite_properties = tuple(sorted(readwrite_properties))
        special_methods = tuple(sorted(special_methods))
        static_methods = tuple(sorted(static_methods))
        result = (
            class_methods,
            data,
            methods,
            readonly_properties,
            readwrite_properties,
            special_methods,
            static_methods,
        )
        return result


class SummarizingModuleDocumenter(ModuleDocumenter):
    """
    A summarizing module documenter.

    Organizes member documenters by their *documentation section*.

    Treats *nominative* submodule documenters (almost) as if they were locally
    defined members. This means that a documenter for a package which contains
    a module which contains *only* one class or function named identically to
    that module will treat that class or function as though it was defined in
    the package's ``__init__.py``. These nominative submodule members are
    organized in each documentation section via an *autosummary* table, rather
    than including their documentation directly.

    ::

        >>> import uqbar.apis
        >>> documenter = uqbar.apis.SummarizingModuleDocumenter("uqbar.io")
        >>> print(str(documenter))
        .. _uqbar--io:
        <BLANKLINE>
        io
        ==
        <BLANKLINE>
        .. automodule:: uqbar.io
        <BLANKLINE>
        .. currentmodule:: uqbar.io
        <BLANKLINE>
        .. container:: svg-container
        <BLANKLINE>
           .. inheritance-diagram:: uqbar
              :lineage: uqbar.io
        <BLANKLINE>
        .. raw:: html
        <BLANKLINE>
           <hr/>
        <BLANKLINE>
        .. rubric:: Classes
           :class: section-header
        <BLANKLINE>
        .. autosummary::
           :nosignatures:
        <BLANKLINE>
           ~DirectoryChange
           ~Profiler
           ~RedirectedStreams
           ~Timer
        <BLANKLINE>
        .. autoclass:: DirectoryChange
           :members:
           :undoc-members:
        <BLANKLINE>
        .. autoclass:: Profiler
           :members:
           :undoc-members:
        <BLANKLINE>
        .. autoclass:: RedirectedStreams
           :members:
           :undoc-members:
        <BLANKLINE>
        .. autoclass:: Timer
           :members:
           :undoc-members:
        <BLANKLINE>
        .. raw:: html
        <BLANKLINE>
           <hr/>
        <BLANKLINE>
        .. rubric:: Functions
           :class: section-header
        <BLANKLINE>
        .. autosummary::
           :nosignatures:
        <BLANKLINE>
           ~find_common_prefix
           ~find_executable
           ~open_path
           ~relative_to
           ~walk
           ~write
        <BLANKLINE>
        .. autofunction:: find_common_prefix
        <BLANKLINE>
        .. autofunction:: find_executable
        <BLANKLINE>
        .. autofunction:: open_path
        <BLANKLINE>
        .. autofunction:: relative_to
        <BLANKLINE>
        .. autofunction:: walk
        <BLANKLINE>
        .. autofunction:: write

    :param package_path: the module path of the module to document
    :param document_private_members: whether to documenter private module members
    :param member_documenter_classes: a list of
        :py:class:`~uqbar.apis.MemberDocumenter` subclasses, defining what classes
        to use to identify and document module members
    :param module_documenters: a list of of documenters for submodules and
        subpackages of the documented module; these are generated by an
        :py:class:`~uqbar.apis.APIBuilder` instance rather than the module
        documenter directly
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Documenters"

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        result = self._build_preamble()
        package_path = self.package_path.partition(".")[0]
        lineage_path = self.package_path
        result.extend(
            [
                "",
                ".. container:: svg-container",
                "",
                "   .. inheritance-diagram:: {}".format(package_path),
                "      :lineage: {}".format(lineage_path),
            ]
        )
        if self.is_nominative:
            result.extend(["", str(self.member_documenters[0])])
        else:
            if self.is_package:
                subpackage_documenters = [
                    _
                    for _ in self.module_documenters or []
                    if _.is_package or not _.is_nominative
                ]
                if subpackage_documenters:
                    result.extend(
                        [
                            "",
                            ".. raw:: html",
                            "",
                            "   <hr/>",
                            "",
                            ".. rubric:: Subpackages",
                            "   :class: section-header",
                        ]
                    )
                    result.extend(
                        self._build_toc(subpackage_documenters, show_full_paths=True)
                    )
            for section, documenters in self.member_documenters_by_section:
                result.extend(
                    [
                        "",
                        ".. raw:: html",
                        "",
                        "   <hr/>",
                        "",
                        ".. rubric:: {}".format(section),
                        "   :class: section-header",
                    ]
                )
                local_documenters = [
                    documenter
                    for documenter in documenters
                    if documenter.client.__module__ == self.package_path
                ]
                result.extend(self._build_toc(documenters))
                for local_documenter in local_documenters:
                    result.extend(["", str(local_documenter)])
        return "\n".join(result)

    ### PRIVATE METHODS ###

    def _build_toc(
        self, documenters, show_full_paths: bool = False, **kwargs
    ) -> List[str]:
        result: List[str] = []
        if not documenters:
            return result
        toctree_paths = set()
        for documenter in documenters:
            path = self._build_toc_path(documenter)
            if path:
                toctree_paths.add(path)
        if toctree_paths:
            result.extend(["", ".. toctree::", "   :hidden:", ""])
            for toctree_path in sorted(toctree_paths):
                result.append("   {}".format(toctree_path))
        result.extend(["", ".. autosummary::", "   :nosignatures:", ""])
        for documenter in documenters:
            template = "   ~{}"
            if show_full_paths:
                template = "   {}"
            path = documenter.package_path.rpartition(self.package_path + ".")[-1]
            result.append(template.format(path))
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def member_documenters_by_section(self) -> List[Tuple[str, List[MemberDocumenter]]]:
        result: MutableMapping[str, List[MemberDocumenter]] = {}
        for documenter in self.member_documenters:
            result.setdefault(documenter.documentation_section, []).append(documenter)
        for module_documenter in self.module_documenters or []:
            if not module_documenter.is_nominative:
                continue
            documenter = module_documenter.member_documenters[0]
            result.setdefault(documenter.documentation_section, []).append(documenter)
        return sorted(result.items())


class SummarizingRootDocumenter(RootDocumenter):
    """
    A summarizing root documenter.

    Like the :py:class:`~uqbar.apis.SummarizingClassDocumenter` and
    :py:class:`~uqbar.apis.SummarizingModuleDocumenter`, this subclass of
    :py:class:`~uqbar.apis.RootDocumenter` generates root API documentation
    with additional information and organization.

    Each visited module or package receives its own section with members
    organized by their documentation sections and listed in autosummary tables.

    ::

        >>> import uqbar.apis
        >>> documenter = uqbar.apis.SummarizingRootDocumenter(
        ...     module_documenters=[
        ...         uqbar.apis.ModuleDocumenter('uqbar.io'),
        ...         uqbar.apis.ModuleDocumenter('uqbar.strings'),
        ...         ],
        ...     )
        >>> print(str(documenter))
        API
        ===
        <BLANKLINE>
        .. toctree::
           :hidden:
        <BLANKLINE>
           uqbar/io
           uqbar/strings
        <BLANKLINE>
        .. raw:: html
        <BLANKLINE>
           <hr/>
        <BLANKLINE>
        .. rubric:: :ref:`uqbar.io <uqbar--io>`
           :class: section-header
        <BLANKLINE>
        Tools for IO and file-system manipulation.
        <BLANKLINE>
        .. raw:: html
        <BLANKLINE>
           <hr/>
        <BLANKLINE>
        .. rubric:: Classes
           :class: subsection-header
        <BLANKLINE>
        .. autosummary::
           :nosignatures:
        <BLANKLINE>
           ~uqbar.io.DirectoryChange
           ~uqbar.io.Profiler
           ~uqbar.io.RedirectedStreams
           ~uqbar.io.Timer
        <BLANKLINE>
        .. raw:: html
        <BLANKLINE>
           <hr/>
        <BLANKLINE>
        .. rubric:: Functions
           :class: subsection-header
        <BLANKLINE>
        .. autosummary::
           :nosignatures:
        <BLANKLINE>
           ~uqbar.io.find_common_prefix
           ~uqbar.io.find_executable
           ~uqbar.io.open_path
           ~uqbar.io.relative_to
           ~uqbar.io.walk
           ~uqbar.io.write
        <BLANKLINE>
        .. raw:: html
        <BLANKLINE>
           <hr/>
        <BLANKLINE>
        .. rubric:: :ref:`uqbar.strings <uqbar--strings>`
           :class: section-header
        <BLANKLINE>
        Tools for string manipulation.
        <BLANKLINE>
        .. raw:: html
        <BLANKLINE>
           <hr/>
        <BLANKLINE>
        .. rubric:: Functions
           :class: subsection-header
        <BLANKLINE>
        .. autosummary::
           :nosignatures:
        <BLANKLINE>
           ~uqbar.strings.ansi_escape
           ~uqbar.strings.delimit_words
           ~uqbar.strings.normalize
           ~uqbar.strings.to_dash_case
           ~uqbar.strings.to_snake_case

    :param module_documenters: a list of of documenters for modules and
        packages of the root documenter; these are generated by an
        :py:class:`~uqbar.apis.APIBuilder` instance rather than the module
        documenter directly
    """

    def __str__(self):
        result = [
            self.title,
            "=" * len(self.title),
            "",
            ".. toctree::",
            "   :hidden:",
            "",
        ]
        for documenter in self.module_documenters:
            path = documenter.package_path.replace(".", "/")
            if documenter.is_package:
                path += "/index"
            result.append("   {}".format(path))
        for module_documenter, documenters_by_section in self._recurse(self):
            result.extend(
                [
                    "",
                    ".. raw:: html",
                    "",
                    "   <hr/>",
                    "",
                    ".. rubric:: :ref:`{} <{}>`".format(
                        module_documenter.package_path, module_documenter.reference_name
                    ),
                    "   :class: section-header",
                ]
            )
            summary = self._extract_summary(module_documenter)
            if summary:
                result.extend(["", summary])
            for section_name, documenters in documenters_by_section:
                result.extend(
                    [
                        "",
                        ".. raw:: html",
                        "",
                        "   <hr/>",
                        "",
                        ".. rubric:: {}".format(section_name),
                        "   :class: subsection-header",
                        "",
                        ".. autosummary::",
                        "   :nosignatures:",
                        "",
                    ]
                )
                for documenter in documenters:
                    result.append("   ~{}".format(documenter.package_path))
        return "\n".join(result)

    def _recurse(self, documenter):
        result = []
        if isinstance(documenter, ModuleDocumenter) and not documenter.is_nominative:
            result.append((documenter, documenter.member_documenters_by_section))
        for module_documenter in documenter.module_documenters:
            result.extend(self._recurse(module_documenter))
        return result

    @classmethod
    def _extract_summary(cls, documenter):
        lines = normalize(documenter.client.__doc__ or "").splitlines()
        if not lines:
            return
        settings = mock.Mock(
            auto_id_prefix="",
            id_prefix="",
            language_code="en",
            pep_reference=False,
            rfc_reference=False,
        )
        document = new_document("", settings)
        try:
            summary = extract_summary(lines, document)
        except Exception:
            return ""
        return "\n".join(textwrap.wrap(summary, 79))


class ImmaterialClassDocumenter(SummarizingClassDocumenter):
    """
    Class documenter that plays well with sphinx-immaterial theme.
    """

    ignored_special_methods: Tuple[str, ...] = (
        "__delattr__",
        "__dict__",
        "__eq__",
        "__getattribute__",
        "__getnewargs__",
        "__getstate__",
        "__hash__",
        "__init__",
        "__new__",
        "__postinit__",
        "__reduce__",
        "__reduce_ex__",
        "__repr__",
        "__setattr__",
        "__setstate__",
        "__sizeof__",
        "__str__",
        "__subclasshook__",
        "fromkeys",
        "pipe_cloexec",
    )

    def __str__(self) -> str:
        name = getattr(self.client, "__name__")
        if issubclass(self.client, Exception):  # type: ignore
            return ".. autoexception:: {}".format(name)
        result = [".. autoclass:: {}".format(name), "   :show-inheritance:"]
        if issubclass(self.client, enum.Enum):  # type: ignore
            result.extend(["   :members:", "   :undoc-members:"])
        result.append("")
        attrs: List[str] = []
        for attr in sorted(
            inspect.classify_class_attrs(cast(type, self.client)), key=lambda x: x.name
        ):
            if attr.defining_class is not self.client:
                continue
            if attr.name.startswith("_") and not attr.name.startswith("__"):
                continue
            if attr.name.startswith("__") and attr.name in self.ignored_special_methods:
                continue
            if attr.kind in ["method", "class method", "static method"]:
                attrs.append(f"   .. automethod:: {attr.name}")
            elif attr.kind in ["property"]:
                attrs.append(f"   .. autoproperty:: {attr.name}")
            elif attr.kind in ["data"] and isinstance(attr.object, type):
                attrs.append(
                    f"   .. autoclass:: {self.client.__module__}::{name}.{attr.name}"
                )
        result.extend(sorted(attrs))
        return "\n".join(result)


class ImmaterialModuleDocumenter(ModuleDocumenter):
    """
    Module documenter that plays well with sphinx-immaterial theme.
    """

    def _build_toc(self, documenters, **kwargs) -> List[str]:
        result: List[str] = []
        if not documenters:
            return result
        result.extend(["", ".. toctree::", "   :hidden:"])
        result.append("")
        module_documenters = [_ for _ in documenters if isinstance(_, type(self))]
        for module_documenter in module_documenters:
            path = self._build_toc_path(module_documenter)
            if path:
                result.append("   {}".format(path))
        return result



from uqbar.graphs import Graph, Node, RecordField, RecordGroup
from uqbar.strings import normalize


def test_graphs_RecordField():
    graph = Graph()
    graph.node_attributes["shape"] = "record"

    node_1_field_0 = RecordField(label="left")
    node_1_field_1 = RecordField(label="middle")
    node_1_field_2 = RecordField(label="right")

    node_1 = Node()
    node_1.extend([node_1_field_0, node_1_field_1, node_1_field_2])

    node_2_field_0 = RecordField(label="one")
    node_2_field_1 = RecordField(label="two")

    node_2 = Node()
    node_2.extend([node_2_field_0, node_2_field_1])

    node_3_field_hello = RecordField(label="hello")
    node_3_field_b = RecordField(label="b")
    node_3_field_c = RecordField(label="c")
    node_3_field_d = RecordField(label="d")
    node_3_field_e = RecordField(label="e")
    node_3_field_f = RecordField(label="f")
    node_3_field_g = RecordField(label="g")
    node_3_field_h = RecordField(label="h")

    inner_group = RecordGroup()
    inner_group.extend([node_3_field_c, node_3_field_d, node_3_field_e])

    outer_group = RecordGroup()
    outer_group.extend([node_3_field_b, inner_group, node_3_field_f])

    node_3 = Node()
    node_3.extend([node_3_field_hello, outer_group, node_3_field_g, node_3_field_h])

    graph.extend([node_1, node_2, node_3])

    node_1_field_1.attach(node_2_field_0)
    node_1_field_2.attach(node_3_field_d)

    graphviz_format = format(graph, "graphviz")

    assert graphviz_format == normalize(
        """
        digraph G {
            node [shape=record];
            node_0 [label="<f_0> left | <f_1> middle | <f_2> right"];
            node_1 [label="<f_0> one | <f_1> two"];
            node_2 [label="<f_0> hello | { <f_1_0> b | { <f_1_1_0> c | <f_1_1_1> d | <f_1_1_2> e } | <f_1_2> f } | <f_2> g | <f_3> h"];
            node_0:f_1 -> node_1:f_0;
            node_0:f_2 -> node_2:f_1_1_1;
        }
        """
    )



import unittest

import uqbar.graphs
from uqbar.strings import normalize


class TestCase(unittest.TestCase):
    def test___init__(self):
        node_a = uqbar.graphs.Node(name="foo")
        node_b = uqbar.graphs.Node(name="bar")
        uqbar.graphs.Edge().attach(node_a, node_b)

    def test___format___str(self):
        node_a = uqbar.graphs.Node(name="foo")
        node_b = uqbar.graphs.Node(name="bar")
        edge = uqbar.graphs.Edge().attach(node_a, node_b)
        assert format(edge) == repr(edge)

        node_a = uqbar.graphs.Node(name="foo")
        node_b = uqbar.graphs.Node(name="bar")
        attributes = uqbar.graphs.Attributes(
            mode="edge", color="blue", style=["dotted"]
        )
        edge = uqbar.graphs.Edge(attributes=attributes).attach(node_a, node_b)
        assert format(edge) == repr(edge)

    def test___format___graphviz(self):
        node_a = uqbar.graphs.Node(name="foo")
        node_b = uqbar.graphs.Node(name="bar")
        edge = uqbar.graphs.Edge().attach(node_a, node_b)
        assert format(edge, "graphviz") == "foo -> bar;"

        node_a = uqbar.graphs.Node(name="foo")
        node_b = uqbar.graphs.Node(name="bar")
        attributes = uqbar.graphs.Attributes(
            mode="edge", color="blue", style=["dotted"]
        )
        edge = uqbar.graphs.Edge(attributes=attributes).attach(node_a, node_b)
        assert format(edge, "graphviz") == normalize(
            """
            foo -> bar [color=blue,
                style=dotted];
        """
        )



"""
Uqbar Sphinx inheritance graph extension.

Install by adding ``'uqbar.sphinx.inheritance'`` to the ``extensions`` list in
your Sphinx configuration.

Invoke by adding the following directive to your ReST source:

::

    .. inheritance-graph:: some.python.path

This will generate Graphviz output mediated by a
:py:class:`uqbar.apis.InheritanceGraph` instance. You can specify lineage paths
by adding a ``:lineage:`` option to the directive:

::

    .. inheritance-graph:: some.python.path
       :lineage: some.lineage.path

Both the main argument to the directive and the lineage option can take
multiple paths, separated by spaces:

::

    .. inheritance-graph:: path.one path.two
       :lineage: path.three path.four path.five

"""

import math
import os
import pathlib
import subprocess
from typing import Any, Dict, List, Mapping, Optional, cast

from docutils.nodes import Element, General, Node, SkipNode
from docutils.parsers.rst import Directive, directives
from sphinx.ext.graphviz import graphviz, render_dot_html, render_dot_latex
from sphinx.writers.html import HTMLTranslator
from sphinx.writers.latex import LaTeXTranslator

import uqbar.apis


class inheritance_diagram(General, Element):
    """
    A docutils node to use as a placeholder for the inheritance diagram.
    """

    __documentation_ignore_inherited__ = True


class InheritanceDiagram(Directive):
    """
    Runs when the inheritance_diagram directive is first encountered.
    """

    has_content = False
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {"lineage": directives.unchanged}

    __documentation_ignore_inherited__ = True

    def run(self) -> List[Node]:
        node = inheritance_diagram()
        node.document = self.state.document
        package_paths = self.arguments[0].split()
        lineage_paths = self.options.get("lineage", "").split() or None
        # Create the graph
        try:
            graph = uqbar.apis.InheritanceGraph(
                package_paths=package_paths, lineage_paths=lineage_paths
            )
        except Exception as error:
            if node.document is not None:
                warning = node.document.reporter.warning(
                    error.args[0], line=self.lineno
                )
            return [warning]
        if not graph.all_class_paths:
            return []
        # Create xref nodes for each target of the graph's image map and add
        # them to the doc tree so that Sphinx can resolve the references to
        # real URLs later.  These nodes will eventually be removed from the
        # doctree after we're done with them.
        env = self.state.document.settings.env
        class_role = env.get_domain("py").role("class")
        for class_path in graph.all_class_paths:
            module_name, _, class_name = class_path.rpartition(".")
            url_name = class_name
            if module_name not in ("__builtins__", "builtins"):
                url_name = class_path
            refnodes, _ = class_role(
                "class", ":class:`{}`".format(url_name), url_name, 0, self.state
            )
            node.extend(refnodes)
        # Store the graph object so we can use it to generate the dot file
        # later on
        node["graph"] = graph
        return [node]


def build_urls(self: HTMLTranslator, node: inheritance_diagram) -> Mapping[str, str]:
    """
    Builds a mapping of class paths to URLs.
    """
    current_filename = self.builder.current_docname + self.builder.out_suffix
    urls: Dict[str, str] = {}
    for child in node:
        if not isinstance(child, Element):
            continue
        # Another document
        refuri: Optional[str]
        if (refuri := child.attributes.get("refuri")) is not None:
            package_path: str = child["reftitle"]
            if refuri.startswith("http"):
                _, _, package_path = refuri.partition("#")
            else:
                refuri = str(
                    pathlib.Path("..")
                    / pathlib.Path(current_filename).parent
                    / pathlib.Path(refuri)
                ).replace(os.path.sep, "/")
            urls[package_path] = refuri
        # Same document
        elif (refid := child.get("refid")) is not None:
            urls[child.attributes["reftitle"]] = "../" + current_filename + "#" + refid
    return urls


def html_visit_inheritance_diagram(
    self: HTMLTranslator, node: inheritance_diagram
) -> None:
    """
    Builds HTML output from an :py:class:`~uqbar.sphinx.inheritance.inheritance_diagram` node.
    """
    inheritance_graph = node["graph"]
    urls = build_urls(self, node)
    graphviz_graph = inheritance_graph.build_graph(urls)
    dot_code = format(graphviz_graph, "graphviz")
    # TODO: We can perform unflattening here
    aspect_ratio = inheritance_graph.aspect_ratio
    if aspect_ratio:
        aspect_ratio = math.ceil(math.sqrt(aspect_ratio[1] / aspect_ratio[0]))
    if aspect_ratio > 1:
        process = subprocess.Popen(
            ["unflatten", "-l", str(aspect_ratio), "-c", str(aspect_ratio), "-f"],
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = process.communicate(dot_code.encode())
        dot_code = stdout.decode()
    render_dot_html(
        self, cast(graphviz, node), dot_code, {}, "inheritance", "inheritance"
    )
    raise SkipNode


def latex_visit_inheritance_diagram(
    self: LaTeXTranslator, node: inheritance_diagram
) -> None:
    """
    Builds LaTeX output from an :py:class:`~uqbar.sphinx.inheritance.inheritance_diagram` node.
    """
    inheritance_graph = node["graph"]
    graphviz_graph = inheritance_graph.build_graph()
    graphviz_graph.attributes["size"] = 6.0
    dot_code = format(graphviz_graph, "graphviz")
    render_dot_latex(self, cast(graphviz, node), dot_code, {}, "inheritance")
    raise SkipNode


def skip(self, node: inheritance_diagram) -> None:
    """
    Skip generating output, for non-supported builders.
    """
    raise SkipNode


def setup(app) -> Dict[str, Any]:
    """
    Sets up Sphinx extension.
    """
    app.setup_extension("sphinx.ext.graphviz")
    app.add_node(
        inheritance_diagram,
        html=(html_visit_inheritance_diagram, None),
        latex=(latex_visit_inheritance_diagram, None),
        man=(skip, None),
        texinfo=(skip, None),
        text=(skip, None),
    )
    app.add_directive("inheritance-diagram", InheritanceDiagram)
    return {
        "version": uqbar.__version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from typing import Optional, Tuple

from ..containers import UniqueTreeList, UniqueTreeNode
from .core import Attachable


class RecordField(Attachable, UniqueTreeNode):
    """
    A Graphviz record field.

    ::

        >>> import uqbar.graphs
        >>> field = uqbar.graphs.RecordField(label='My Label')
        >>> print(format(field, 'graphviz'))
        <f_0> My Label

    Port names are generated based on the graph order of the field within its
    parent node:

    ::

        >>> other_field = uqbar.graphs.RecordField(label='Other Label')
        >>> group = uqbar.graphs.RecordGroup()
        >>> node = uqbar.graphs.Node(attributes=dict(shape='Mrecord'))
        >>> cluster = uqbar.graphs.Graph(is_cluster=True)
        >>> graph = uqbar.graphs.Graph()

    ::

        >>> graph.append(cluster)
        >>> cluster.append(node)
        >>> node.append(group)
        >>> group.extend([other_field, field])
        >>> print(format(field, 'graphviz'))
        <f_0_1> My Label

    If a node contains record fields or groups, their format contributions
    override any string label in the node's attributes:

    ::

        >>> print(format(node, 'graphviz'))
        node_0_0 [label="{ <f_0_0> Other Label | <f_0_1> My Label }",
            shape=Mrecord];

    ::

        >>> node.attributes['label'] = 'Foo'
        >>> print(format(node, 'graphviz'))
        node_0_0 [label="{ <f_0_0> Other Label | <f_0_1> My Label }",
            shape=Mrecord];

    ::

        >>> node[:] = []
        >>> print(format(node, 'graphviz'))
        node_0_0 [label=Foo,
            shape=Mrecord];

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Record Field Classes"

    ### INITIALIZER ###

    def __init__(
        self, label: Optional[str] = None, *, name: Optional[str] = None
    ) -> None:
        UniqueTreeNode.__init__(self, name=name)
        Attachable.__init__(self)
        if label is not None:
            label = str(label)
        self._label = label

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: Optional[str] = None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == "graphviz":
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        result = "<{}>".format(self._get_port_name())
        if self.label:
            result = "{} {}".format(result, self.label)
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def label(self) -> Optional[str]:
        return self._label


class RecordGroup(UniqueTreeList):
    """
    A Graphviz record field group.

    ::

        >>> import uqbar.graphs
        >>> group = uqbar.graphs.RecordGroup()
        >>> group.extend([
        ...     uqbar.graphs.RecordField(),
        ...     uqbar.graphs.RecordGroup([
        ...         uqbar.graphs.RecordField(),
        ...         uqbar.graphs.RecordField(),
        ...         ]),
        ...     uqbar.graphs.RecordField(),
        ...     ])
        >>> print(format(group, 'graphviz'))
        { <f_0> | { <f_0> | <f_0> } | <f_0> }

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Record Field Classes"

    ### INITIALIZER ###

    def __init__(self, children=None, *, name: Optional[str] = None) -> None:
        UniqueTreeList.__init__(self, name=name, children=children)

    ### SPECIAL METHODS ###

    def __format__(self, format_spec=None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == "graphviz":
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        result = " | ".join(_ for _ in (format(_, "graphviz") for _ in self) if _)
        if result:
            result = "{{ {} }}".format(result)
        return result

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self) -> Tuple[type, ...]:
        import uqbar.graphs

        return (uqbar.graphs.RecordField, uqbar.graphs.RecordGroup)



"""
Specialized container classes.
"""

from .dependency_graph import DependencyGraph
from .unique_tree import (
    UniqueTreeDict,
    UniqueTreeList,
    UniqueTreeNode,
    UniqueTreeSet,
    UniqueTreeTuple,
)

__all__ = [
    "DependencyGraph",
    "UniqueTreeDict",
    "UniqueTreeList",
    "UniqueTreeNode",
    "UniqueTreeSet",
    "UniqueTreeTuple",
]



import os
import pathlib
import shutil
import sys

import pytest

import uqbar.apis
from uqbar.strings import normalize


@pytest.fixture
def test_path():
    test_path = pathlib.Path(__file__).parent
    docs_path = test_path / "docs"
    if str(test_path) not in sys.path:
        sys.path.insert(0, str(test_path))
    if docs_path.exists():
        shutil.rmtree(str(docs_path))
    yield test_path
    if docs_path.exists():
        shutil.rmtree(str(docs_path))


def test_collection_01(test_path):
    builder = uqbar.apis.APIBuilder([test_path / "fake_package"], test_path / "docs")
    source_paths = uqbar.apis.collect_source_paths(builder._initial_source_paths)
    node_tree = builder.build_node_tree(source_paths)
    assert normalize(str(node_tree)) == normalize(
        """
        None/
            fake_package/
                fake_package.empty_module
                fake_package.empty_package/
                    fake_package.empty_package.empty
                fake_package.enums
                fake_package.module
                fake_package.multi/
                    fake_package.multi.one
                    fake_package.multi.two
        """
    )
    documenters = list(builder.collect_module_documenters(node_tree))
    assert isinstance(documenters[0], uqbar.apis.RootDocumenter)
    assert [documenter.package_path for documenter in documenters[1:]] == [
        "fake_package",
        "fake_package.empty_module",
        "fake_package.empty_package",
        "fake_package.empty_package.empty",
        "fake_package.enums",
        "fake_package.module",
        "fake_package.multi",
        "fake_package.multi.one",
        "fake_package.multi.two",
    ]


def test_collection_02(test_path):
    builder = uqbar.apis.APIBuilder(
        [test_path / "fake_package"], test_path / "docs", document_private_modules=True
    )
    source_paths = uqbar.apis.collect_source_paths(builder._initial_source_paths)
    node_tree = builder.build_node_tree(source_paths)
    assert normalize(str(node_tree)) == normalize(
        """
        None/
            fake_package/
                fake_package._private/
                    fake_package._private.nested
                fake_package.empty_module
                fake_package.empty_package/
                    fake_package.empty_package.empty
                fake_package.enums
                fake_package.module
                fake_package.multi/
                    fake_package.multi.one
                    fake_package.multi.two
        """
    )
    documenters = list(builder.collect_module_documenters(node_tree))
    assert isinstance(documenters[0], uqbar.apis.RootDocumenter)
    assert [documenter.package_path for documenter in documenters[1:]] == [
        "fake_package",
        "fake_package._private",
        "fake_package._private.nested",
        "fake_package.empty_module",
        "fake_package.empty_package",
        "fake_package.empty_package.empty",
        "fake_package.enums",
        "fake_package.module",
        "fake_package.multi",
        "fake_package.multi.one",
        "fake_package.multi.two",
    ]


def test_collection_03(test_path):
    builder = uqbar.apis.APIBuilder(
        [test_path / "fake_package" / "multi"], test_path / "docs"
    )
    source_paths = uqbar.apis.collect_source_paths(builder._initial_source_paths)
    node_tree = builder.build_node_tree(source_paths)
    documenters = list(builder.collect_module_documenters(node_tree))
    assert isinstance(documenters[0], uqbar.apis.RootDocumenter)
    assert [documenter.package_path for documenter in documenters[1:]] == [
        "fake_package",
        "fake_package.multi",
        "fake_package.multi.one",
        "fake_package.multi.two",
    ]


def test_collection_04(test_path):
    builder = uqbar.apis.APIBuilder(
        [test_path / "fake_package"], test_path / "docs", document_empty_modules=False
    )
    source_paths = uqbar.apis.collect_source_paths(builder._initial_source_paths)
    node_tree = builder.build_node_tree(source_paths)
    assert normalize(str(node_tree)) == normalize(
        """
        None/
            fake_package/
                fake_package.enums
                fake_package.module
                fake_package.multi/
                    fake_package.multi.one
                    fake_package.multi.two
        """
    )
    documenters = list(builder.collect_module_documenters(node_tree))
    assert isinstance(documenters[0], uqbar.apis.RootDocumenter)
    assert [documenter.package_path for documenter in documenters[1:]] == [
        "fake_package",
        "fake_package.enums",
        "fake_package.module",
        "fake_package.multi",
        "fake_package.multi.one",
        "fake_package.multi.two",
    ]


def test_output_01(test_path):
    builder = uqbar.apis.APIBuilder([test_path / "fake_package"], test_path / "docs")
    builder()
    paths = sorted((test_path / "docs").rglob("*"))
    paths = [
        str(path.relative_to(test_path)).replace(os.path.sep, "/") for path in paths
    ]
    assert paths == [
        "docs/fake_package",
        "docs/fake_package/empty_module.rst",
        "docs/fake_package/empty_package",
        "docs/fake_package/empty_package/empty.rst",
        "docs/fake_package/empty_package/index.rst",
        "docs/fake_package/enums.rst",
        "docs/fake_package/index.rst",
        "docs/fake_package/module.rst",
        "docs/fake_package/multi",
        "docs/fake_package/multi/index.rst",
        "docs/fake_package/multi/one.rst",
        "docs/fake_package/multi/two.rst",
        "docs/index.rst",
    ]
    base_path = test_path / "docs" / "fake_package"
    with (base_path / ".." / "index.rst").open() as file_pointer:
        assert normalize(file_pointer.read()) == normalize(
            """
            API
            ===

            .. toctree::

               fake_package/index
            """
        )
    with (base_path / "index.rst").open() as file_pointer:
        assert normalize(file_pointer.read()) == normalize(
            """
            .. _fake-package:

            fake_package
            ============

            .. automodule:: fake_package

            .. currentmodule:: fake_package

            .. toctree::

               empty_module
               empty_package/index
               enums
               module
               multi/index
            """
        )
    with (base_path / "module.rst").open() as file_pointer:
        assert normalize(file_pointer.read()) == normalize(
            """
            .. _fake-package--module:

            module
            ======

            .. automodule:: fake_package.module

            .. currentmodule:: fake_package.module

            .. autoclass:: ChildClass
               :members:
               :undoc-members:

            .. autoclass:: PublicClass
               :members:
               :undoc-members:

            .. autofunction:: public_function
            """
        )


def test_output_02(test_path):
    builder = uqbar.apis.APIBuilder(
        [test_path / "fake_package"], test_path / "docs", document_private_modules=True
    )
    builder()
    paths = sorted((test_path / "docs").rglob("*"))
    paths = [
        str(path.relative_to(test_path)).replace(os.path.sep, "/") for path in paths
    ]
    assert paths == [
        "docs/fake_package",
        "docs/fake_package/_private",
        "docs/fake_package/_private/index.rst",
        "docs/fake_package/_private/nested.rst",
        "docs/fake_package/empty_module.rst",
        "docs/fake_package/empty_package",
        "docs/fake_package/empty_package/empty.rst",
        "docs/fake_package/empty_package/index.rst",
        "docs/fake_package/enums.rst",
        "docs/fake_package/index.rst",
        "docs/fake_package/module.rst",
        "docs/fake_package/multi",
        "docs/fake_package/multi/index.rst",
        "docs/fake_package/multi/one.rst",
        "docs/fake_package/multi/two.rst",
        "docs/index.rst",
    ]
    base_path = test_path / "docs" / "fake_package"
    with (base_path / ".." / "index.rst").open() as file_pointer:
        assert normalize(file_pointer.read()) == normalize(
            """
            API
            ===

            .. toctree::

               fake_package/index
            """
        )
    with (base_path / "index.rst").open() as file_pointer:
        assert normalize(file_pointer.read()) == normalize(
            """
            .. _fake-package:

            fake_package
            ============

            .. automodule:: fake_package

            .. currentmodule:: fake_package

            .. toctree::

               _private/index
               empty_module
               empty_package/index
               enums
               module
               multi/index
            """
        )
    with (base_path / "module.rst").open() as file_pointer:
        assert normalize(file_pointer.read()) == normalize(
            """
            .. _fake-package--module:

            module
            ======

            .. automodule:: fake_package.module

            .. currentmodule:: fake_package.module

            .. autoclass:: ChildClass
               :members:
               :undoc-members:

            .. autoclass:: PublicClass
               :members:
               :undoc-members:

            .. autofunction:: public_function
            """
        )


def test_output_03(test_path):
    builder = uqbar.apis.APIBuilder(
        [test_path / "fake_package"],
        test_path / "docs",
        document_private_members=True,
        document_private_modules=True,
    )
    builder()
    paths = sorted((test_path / "docs").rglob("*"))
    paths = [
        str(path.relative_to(test_path)).replace(os.path.sep, "/") for path in paths
    ]
    assert paths == [
        "docs/fake_package",
        "docs/fake_package/_private",
        "docs/fake_package/_private/index.rst",
        "docs/fake_package/_private/nested.rst",
        "docs/fake_package/empty_module.rst",
        "docs/fake_package/empty_package",
        "docs/fake_package/empty_package/empty.rst",
        "docs/fake_package/empty_package/index.rst",
        "docs/fake_package/enums.rst",
        "docs/fake_package/index.rst",
        "docs/fake_package/module.rst",
        "docs/fake_package/multi",
        "docs/fake_package/multi/index.rst",
        "docs/fake_package/multi/one.rst",
        "docs/fake_package/multi/two.rst",
        "docs/index.rst",
    ]
    base_path = test_path / "docs" / "fake_package"
    with (base_path / ".." / "index.rst").open() as file_pointer:
        assert normalize(file_pointer.read()) == normalize(
            """
            API
            ===

            .. toctree::

               fake_package/index
            """
        )
    with (base_path / "index.rst").open() as file_pointer:
        assert normalize(file_pointer.read()) == normalize(
            """
            .. _fake-package:

            fake_package
            ============

            .. automodule:: fake_package

            .. currentmodule:: fake_package

            .. toctree::

               _private/index
               empty_module
               empty_package/index
               enums
               module
               multi/index
            """
        )
    with (base_path / "module.rst").open() as file_pointer:
        assert normalize(file_pointer.read()) == normalize(
            """
            .. _fake-package--module:

            module
            ======

            .. automodule:: fake_package.module

            .. currentmodule:: fake_package.module

            .. autoclass:: ChildClass
               :members:
               :undoc-members:

            .. autoclass:: PublicClass
               :members:
               :undoc-members:

            .. autoclass:: _PrivateClass
               :members:
               :undoc-members:

            .. autofunction:: _private_function

            .. autofunction:: public_function
            """
        )



import datetime
import hashlib
import pathlib
import re
import shutil
import subprocess
import tempfile
from typing import Generator, Sequence, Tuple

from ..io import Timer, open_path


class Grapher:
    ### CLASS VARIABLES ###

    _valid_formats = ("png", "pdf", "svg")
    _valid_layouts = ("circo", "dot", "fdp", "neato", "osage", "sfdp", "twopi")

    ### INITIALIZER ###

    def __init__(self, graphable, format_="pdf", layout="dot", output_directory=None):
        if layout not in self._valid_layouts:
            raise ValueError("Invalid layout: {layout!r}")
        if format_ not in self._valid_formats:
            raise ValueError("Invalid format: {format_!r}")
        self.graphable = graphable
        self.format_ = format_
        self.layout = layout
        self.output_directory = pathlib.Path(output_directory or ".")

    ### SPECIAL METHODS ###

    def __call__(self):
        with Timer() as format_timer:
            string = self.get_string()
        format_time = format_timer.elapsed_time
        layout = self.get_layout()
        format_ = self.get_format()
        render_prefix = self.get_render_prefix(string)
        render_directory_path = self.get_render_directory()
        input_path = (render_directory_path / render_prefix).with_suffix(".dot")
        self.persist_string(string, input_path)
        render_command = self.get_render_command(format_, input_path, layout)
        with Timer() as render_timer:
            log, success = self.run_command(render_command)
        render_time = render_timer.elapsed_time
        self.persist_log(log, input_path.with_suffix(".log"))
        output_directory_path = self.get_output_directory()
        output_paths = self.migrate_assets(
            render_prefix, render_directory_path, output_directory_path
        )
        openable_paths = []
        for output_path in self.get_openable_paths(format_, output_paths):
            openable_paths.append(output_path)
            self.open_output_path(output_path)
        return output_path, format_time, render_time, success, log

    ### PUBLIC METHODS ###

    def get_format(self) -> str:
        return self.format_

    def get_layout(self) -> str:
        return self.layout

    def get_openable_paths(
        self, format_, output_paths
    ) -> Generator[pathlib.Path, None, None]:
        for path in output_paths:
            if path.suffix == f".{format_}":
                yield path

    def get_output_directory(self) -> pathlib.Path:
        return self.output_directory

    def get_render_command(self, format_, input_path, layout) -> str:
        parts = [
            layout,
            "-v",
            "-T",
            format_,
            "-o",
            str(input_path.with_suffix("." + format_)),
            str(input_path),
        ]
        return " ".join(parts)

    def get_render_directory(self):
        return pathlib.Path(tempfile.mkdtemp())

    def get_render_prefix(self, string) -> str:
        timestamp = re.sub(r"[^\w]", "-", datetime.datetime.now().isoformat())
        checksum = hashlib.sha1(string.encode()).hexdigest()[:7]
        return f"{timestamp}-{checksum}"

    def get_string(self) -> str:
        try:
            graphviz_graph = self.graphable.__graph__()
            return format(graphviz_graph, "graphviz")
        except AttributeError:
            return self.graphable.__format_graphviz__()

    def migrate_assets(
        self, render_prefix, render_directory, output_directory
    ) -> Sequence[pathlib.Path]:
        migrated_assets = []
        for old_path in render_directory.iterdir():
            if not old_path.name.startswith(render_prefix):
                continue
            new_path = output_directory / old_path.name
            try:
                old_path.rename(new_path)
            except OSError:
                shutil.copy(old_path, new_path)
                old_path.unlink()
            migrated_assets.append(new_path)
        return migrated_assets

    def open_output_path(self, output_path):
        open_path(output_path)

    def persist_log(self, string, input_path):
        input_path.write_text(string)

    def persist_string(self, string, input_path):
        input_path.write_text(string)

    def run_command(self, command: str) -> Tuple[str, int]:
        completed_process = subprocess.run(
            command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        return completed_process.stdout.decode(), completed_process.returncode == 0



import enum
import importlib
import inspect
import pathlib
import types
from typing import List, MutableMapping, Optional, Sequence, Tuple, Type, cast


class MemberDocumenter:
    """
    Abstract base class for module member documenters.

    Member documenters generate reStructuredText output for classes, functions
    and other objects defined in modules. They also implement the logic for
    identifying which of those objects should be selected for documenting via
    their
    :py:meth:`~uqbar.apis.MemberDocumenter.MemberDocumenter.validate_client`
    method. Module documenters loop over their client module's dict and try
    each of their member documenter class' validation method against each dict
    member in turn in order to identify documentable content and associate it
    with a member documenter class.

    .. tip::

       Subclass :py:class:`~uqbar.apis.MemberDocumenter` to implement your own
       custom module member documentation output. You'll need to override
       :py:meth:`~uqbar.apis.MemberDocumenter.MemberDocumenter.validate_client`
       with custom identification logic, and provide your desired
       reStructuredText output via an overridden
       :py:meth:`~uqbar.apis.MemberDocumenter.MemberDocumenter.__str__`
       implementation.

    :param package_path: the module path and name of the member to document
    """

    ### INITIALIZER ###

    def __init__(self, package_path: str) -> None:
        module_path, _, client_name = package_path.rpartition(".")
        module = importlib.import_module(module_path)
        client = getattr(module, client_name)
        if not self.validate_client(client, module_path):
            message = f"Unexpected object: {client!r} from {module_path}"
            raise ValueError(message)
        self._client = client
        self._package_path = package_path

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        raise NotImplementedError

    ### PUBLIC METHODS ###

    @classmethod
    def validate_client(cls, client: object, module_path: str) -> bool:
        """
        Check if the member documenter can document the provided client.

        :param client: the potential client to validate
        :param module_path: the package path of the module where the client was encountered
        """
        raise NotImplementedError

    ### PUBLIC PROPERTIES ###

    @property
    def client(self) -> object:
        return self._client

    @property
    def documentation_section(self) -> str:
        raise NotImplementedError

    @property
    def package_path(self) -> str:
        return self._package_path


class FunctionDocumenter(MemberDocumenter):
    """
    A basic function documenter.

    ::

        >>> import uqbar.apis
        >>> path = 'uqbar.io.walk'
        >>> documenter = uqbar.apis.FunctionDocumenter(path)
        >>> documentation = str(documenter)
        >>> print(documentation)
        .. autofunction:: walk

    .. tip::

       Subclass :py:class:`~uqbar.apis.FunctionDocumenter` to
       implement your own custom module documentation output.
       You'll need to provide your desired reStructuredText output
       via an overridden
       :py:meth:`~uqbar.apis.FunctionDocumenter.FunctionDocumenter.__str__`
       implementation.

    :param package_path: the module path and name of the member to document
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Documenters"

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        return ".. autofunction:: {}".format(getattr(self.client, "__name__"))

    ### PUBLIC METHODS ###

    @classmethod
    def validate_client(cls, client: object, module_path: str) -> bool:
        return (
            isinstance(client, types.FunctionType) and client.__module__ == module_path
        )

    ### PUBLIC PROPERTIES ###

    @property
    def documentation_section(self) -> str:
        return "Functions"


class ClassDocumenter(MemberDocumenter):
    """
    A basic class documenter.

    ::

        >>> import uqbar.apis
        >>> path = 'uqbar.apis.documenters.ClassDocumenter'
        >>> documenter = uqbar.apis.ClassDocumenter(path)
        >>> documentation = str(documenter)
        >>> print(documentation)
        .. autoclass:: ClassDocumenter
           :members:
           :undoc-members:

    .. tip::

       Subclass :py:class:`~uqbar.apis.ClassDocumenter` to implement your own
       custom class documentation output. You'll need to provide your desired
       reStructuredText output via an overridden
       :py:meth:`~uqbar.apis.ClassDocumenter.ClassDocumenter.__str__`
       implementation.

       See :py:class:`~uqbar.apis.documenters.SummarizingClassDocumenter` for an example.

    :param package_path: the module path and name of the member to document
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Documenters"

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        name = getattr(self.client, "__name__")
        if issubclass(self.client, Exception):  # type: ignore
            return ".. autoexception:: {}".format(name)
        return "\n".join(
            [".. autoclass:: {}".format(name), "   :members:", "   :undoc-members:"]
        )

    ### PUBLIC METHODS ###

    @classmethod
    def validate_client(cls, client: object, module_path: str) -> bool:
        return isinstance(client, type) and getattr(client, "__module__") == module_path

    ### PUBLIC PROPERTIES ###

    @property
    def documentation_section(self) -> str:
        client = cast(type, self.client)
        if hasattr(client, "__documentation_section__"):
            section = getattr(client, "__documentation_section__")
            if section is not None:
                return section
        if inspect.isabstract(client):
            return "Abstract Classes"
        elif issubclass(client, enum.Enum):
            return "Enumerations"
        elif issubclass(client, Exception):
            return "Exceptions"
        return "Classes"


class ModuleDocumenter:
    """
    A basic module documenter.

    ::

        >>> import uqbar.apis
        >>> documenter = uqbar.apis.ModuleDocumenter("uqbar.io")
        >>> print(str(documenter))
        .. _uqbar--io:
        <BLANKLINE>
        io
        ==
        <BLANKLINE>
        .. automodule:: uqbar.io
        <BLANKLINE>
        .. currentmodule:: uqbar.io
        <BLANKLINE>
        .. autoclass:: DirectoryChange
           :members:
           :undoc-members:
        <BLANKLINE>
        .. autoclass:: Profiler
           :members:
           :undoc-members:
        <BLANKLINE>
        .. autoclass:: RedirectedStreams
           :members:
           :undoc-members:
        <BLANKLINE>
        .. autoclass:: Timer
           :members:
           :undoc-members:
        <BLANKLINE>
        .. autofunction:: find_common_prefix
        <BLANKLINE>
        .. autofunction:: find_executable
        <BLANKLINE>
        .. autofunction:: open_path
        <BLANKLINE>
        .. autofunction:: relative_to
        <BLANKLINE>
        .. autofunction:: walk
        <BLANKLINE>
        .. autofunction:: write

    .. tip::

       Subclass :py:class:`~uqbar.apis.ModuleDocumenter` to implement your own
       custom module documentation output.
       You'll need to provide your desired reStructuredText output
       via an overridden
       :py:meth:`~uqbar.apis.ModuleDocumenter.ModuleDocumenter.__str__`
       implementation.

       See :py:class:`~uqbar.apis.SummarizingModuleDocumenter` for an example.

    :param package_path: the module path of the module to document
    :param document_private_members: whether to documenter private module members
    :param member_documenter_classes: a list of
        :py:class:`~uqbar.apis.MemberDocumenter` subclasses, defining what classes
        to use to identify and document module members
    :param module_documenters: a list of of documenters for submodules and
        subpackages of the documented module; these are generated by an
        :py:class:`~uqbar.apis.APIBuilder` instance rather than the module
        documenter directly
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Documenters"

    ### INITIALIZER ###

    def __init__(
        self,
        package_path: str,
        document_private_members: bool = False,
        member_documenter_classes: Optional[Sequence[Type[MemberDocumenter]]] = None,
        module_documenters: Optional[Sequence["ModuleDocumenter"]] = None,
    ) -> None:
        self._package_path = package_path
        client = importlib.import_module(package_path)
        assert isinstance(client, types.ModuleType)
        self._client = client
        self._document_private_members = bool(document_private_members)
        if member_documenter_classes is None:
            member_documenter_classes = [ClassDocumenter, FunctionDocumenter]
        for _ in member_documenter_classes:
            assert issubclass(_, MemberDocumenter), _
        self._member_documenter_classes = tuple(member_documenter_classes)
        if module_documenters is not None:
            for submodule_documenter in module_documenters:
                assert isinstance(submodule_documenter, ModuleDocumenter)
            module_documenters = tuple(module_documenters)
        self._module_documenters = module_documenters or ()
        self._member_documenters = self._populate()

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        result = self._build_preamble()
        result.extend(self._build_toc(self.module_documenters or []))
        member_strings: List[str] = []
        for documenter in self._member_documenters:
            member_strings.append(str(documenter))
        member_strings.sort(key=lambda x: x.partition("\n")[0].split("::"))
        for member_string in member_strings:
            result.append("")
            result.append(member_string)
        return "\n".join(result)

    ### PRIVATE METHODS ###

    def _populate(self) -> Sequence[MemberDocumenter]:
        documenters = []
        for name in sorted(dir(self.client)):
            if name.startswith("_") and not self.document_private_members:
                continue
            client = getattr(self.client, name)
            for class_ in self.member_documenter_classes:
                if class_.validate_client(client, self.package_path):
                    path = "{}.{}".format(client.__module__, client.__name__)
                    documenter = class_(path)
                    documenters.append(documenter)
                    break
        return tuple(documenters)

    ### PRIVATE METHODS ###

    def _build_toc(self, documenters, **kwargs) -> List[str]:
        result: List[str] = []
        if not documenters:
            return result
        result.extend(["", ".. toctree::"])
        result.append("")
        module_documenters = [_ for _ in documenters if isinstance(_, type(self))]
        for module_documenter in module_documenters:
            path = self._build_toc_path(module_documenter)
            if path:
                result.append("   {}".format(path))
        return result

    def _build_toc_path(self, documenter):
        path = documenter.package_path.partition(self.package_path + ".")[-1]
        base, _, name = path.rpartition(".")
        if name.lower() == "index":
            path = base + "._" + name
        if not isinstance(documenter, ModuleDocumenter):
            path = path.rpartition(".")[0]
        elif documenter.is_package:
            path += "/index"
        if path.lower() == "index":
            path = "_" + path
        return path

    def _build_preamble(self) -> List[str]:
        result: List[str] = [
            ".. _{}:".format(self.reference_name),
            "",
            self.package_name,
            "=" * len(self.package_name),
            "",
            ".. automodule:: {}".format(self.package_path),
            "",
            ".. currentmodule:: {}".format(self.package_path),
        ]
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def client(self) -> object:
        return self._client

    @property
    def is_package(self) -> bool:
        return hasattr(self.client, "__path__")

    @property
    def document_private_members(self) -> bool:
        return self._document_private_members

    @property
    def documentation_path(self) -> pathlib.Path:
        path = pathlib.Path(".").joinpath(*self.package_path.split("."))
        if self.is_package:
            path = path.joinpath("index")
        elif path.name.lower() == "index":
            name = path.name
            path = path.parent.joinpath("_" + name)
        return path.with_suffix(".rst")

    @property
    def is_nominative(self) -> bool:
        if self.is_package or len(self.member_documenters) != 1:
            return False
        parts = self.member_documenters[0].package_path.split(".")
        return parts[-1] == parts[-2]

    @property
    def member_documenter_classes(self) -> Sequence[Type[MemberDocumenter]]:
        return self._member_documenter_classes

    @property
    def member_documenters(self) -> Sequence[MemberDocumenter]:
        return self._member_documenters

    @property
    def member_documenters_by_section(
        self,
    ) -> Sequence[Tuple[str, Sequence[MemberDocumenter]]]:
        result: MutableMapping[str, List[MemberDocumenter]] = {}
        for documenter in self.member_documenters:
            result.setdefault(documenter.documentation_section, []).append(documenter)
        return sorted(result.items())

    @property
    def module_documenters(self) -> Sequence["ModuleDocumenter"]:
        return self._module_documenters

    @property
    def package_name(self) -> str:
        if "." in self.package_path:
            return self._package_path.rpartition(".")[-1]
        return self._package_path

    @property
    def package_path(self) -> str:
        return self._package_path

    @property
    def reference_name(self) -> str:
        return self.package_path.replace("_", "-").replace(".", "--")


class RootDocumenter:
    """
    A basic root documenter.

    This documenter generates only one reStructuredText document: the root API
    index page, which contains information about all of the modules traversed
    by an :py:class:`~uqbar.apis.APIBuilder`.

    Output is a basic `toctree` directive.

    ::

        >>> import uqbar.apis
        >>> documenter = uqbar.apis.RootDocumenter(
        ...     module_documenters=[
        ...         uqbar.apis.ModuleDocumenter('uqbar.io'),
        ...         uqbar.apis.ModuleDocumenter('uqbar.strings'),
        ...         ],
        ...     )
        >>> print(str(documenter))
        API
        ===
        <BLANKLINE>
        .. toctree::
        <BLANKLINE>
           uqbar/io
           uqbar/strings
        <BLANKLINE>

    .. tip::

       Subclass :py:class:`~uqbar.apis.RootDocumenter` to
       implement your own custom module documentation output.
       You'll need to provide your desired reStructuredText output
       via an overridden
       :py:meth:`~uqbar.apis.RootDocumenter.RootDocumenter.__str__`
       implementation.

       See :py:class:`~uqbar.apis.SummarizingRootDocumenter` for an example.

    :param module_documenters: a list of of documenters for modules and
        packages of the root documenter; these are generated by an
        :py:class:`~uqbar.apis.APIBuilder` instance rather than the module
        documenter directly
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "Documenters"

    ### INITIALIZER ###

    def __init__(self, module_documenters=None, title="API"):
        import uqbar.apis

        if module_documenters is not None:
            assert all(
                isinstance(_, uqbar.apis.ModuleDocumenter) for _ in module_documenters
            ), module_documenters
            module_documenters = tuple(module_documenters)
        self._module_documenters = module_documenters or ()
        self._title = title

    ### SPECIAL METHODS ###

    def __str__(self):
        result = [self.title, "=" * len(self.title), ""]
        if self.module_documenters:
            result.extend([".. toctree::", ""])
            for module_documenter in self.module_documenters:
                path = module_documenter.package_path.replace(".", "/")
                if module_documenter.is_package:
                    path = "{}/index".format(path)
                result.append("   {}".format(path))
            result.append("")
        return "\n".join(result)

    ### PUBLIC PROPERTIES ###

    @property
    def documentation_path(self):
        return pathlib.Path("index.rst")

    @property
    def module_documenters(self):
        return self._module_documenters

    @property
    def title(self):
        return self._title



"""
Tools for IO and file-system manipulation.
"""

import cProfile
import collections
import io
import os
import platform
import pstats
import subprocess
import sys
import time
from pathlib import Path
from typing import Generator, List, Optional, Sequence, Tuple, Union


class DirectoryChange:
    """
    A context manager for temporarily changing the current working directory.

    This context manager is re-entrant.
    """

    ### INITIALIZER ###

    def __init__(self, directory, verbose=False):
        directory = Path(directory).resolve().absolute()
        if not directory.is_dir():
            directory = directory.parent
        self._directory = directory
        self._directory_stack = []
        self._verbose = bool(verbose)

    ### SPECIAL METHODS ###

    def __enter__(self):
        if self.verbose:
            print("Changing directory to {} ...".format(self.directory))
        self._directory_stack.append(Path.cwd())
        os.chdir(str(self._directory))
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        original_directory = self._directory_stack.pop()
        if self.verbose:
            print("Returning to {} ...".format(original_directory))
        os.chdir(str(original_directory))

    ### PUBLIC PROPERTIES ###

    @property
    def directory(self):
        return self._directory

    @property
    def verbose(self):
        return self._verbose


class Profiler:
    """
    A context manager for profiling blocks of code.
    """

    def __enter__(self) -> "Profiler":
        self._profiler = cProfile.Profile()
        self._profiler.enable()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._profiler.disable()
        stream = io.StringIO()
        profiler_stats = pstats.Stats(self._profiler, stream=stream)
        profiler_stats = profiler_stats.sort_stats("cumulative")
        profiler_stats.print_stats()
        print(stream.getvalue())
        profiler_stats.dump_stats("stats.profile")


class RedirectedStreams:
    """
    A context manager for capturing ``stdout`` and ``stderr`` output.

    ..  container:: example

        ::

            >>> import uqbar.io
            >>> import io

        ::

            >>> string_io = io.StringIO()
            >>> with uqbar.io.RedirectedStreams(stdout=string_io):
            ...     print("hello, world!")
            ...

        ::

            >>> result = string_io.getvalue()
            >>> string_io.close()
            >>> print(result)
            hello, world!

    This context manager is not reentrant. Use a separate instance when nesting
    multiple timers.
    """

    ### INITIALIZER ###

    def __init__(self, stdout=None, stderr=None):
        self._stdout = stdout or sys.stdout
        self._stderr = stderr or sys.stderr

    ### SPECIAL METHODS ###

    def __enter__(self):
        self._old_stdout, self._old_stderr = sys.stdout, sys.stderr
        self._old_stdout.flush()
        self._old_stderr.flush()
        sys.stdout, sys.stderr = self._stdout, self._stderr
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self._stdout.flush()
            self._stderr.flush()
        finally:
            sys.stdout = self._old_stdout
            sys.stderr = self._old_stderr

    ### PUBLIC PROPERTIES ###

    @property
    def stderr(self):
        return self._stderr

    @property
    def stdout(self):
        return self._stdout


class Timer:
    """
    A context manager for timing blocks of code.

    This context manager is not reentrant. Use a separate instance when nesting
    multiple timers.

    :param exit_message: message to print on entering the context
    :param enter_message: message to print on exiting the context
    :param verbose: whether to print output

    ::

        >>> import math
        >>> from uqbar.io import Timer
        >>> timer = Timer("Elapsed time:", "Looping!")
        >>> with timer:
        ...     for i in range(10000):
        ...         z = i ** math.pi
        ...
        Looping!
        Elapsed time: 0.0...

    """

    def __init__(
        self,
        exit_message: Optional[str] = None,
        enter_message: Optional[str] = None,
        verbose: bool = True,
    ) -> None:
        if enter_message is not None:
            enter_message = str(enter_message)
        self._enter_message = enter_message
        if exit_message is not None:
            exit_message = str(exit_message)
        self._exit_message = exit_message
        self._start_time: Optional[float] = None
        self._stop_time: Optional[float] = None
        self._verbose = bool(verbose)

    ### SPECIAL METHODS ###

    def __enter__(self) -> "Timer":
        if self.enter_message and self.verbose:
            print(self.enter_message)
        self._stop_time = None
        self._start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._stop_time = time.time()
        if self.exit_message and self.verbose:
            print(self.exit_message, self.elapsed_time)

    ### PUBLIC PROPERTIES ###

    @property
    def elapsed_time(self) -> Union[float, None]:
        if self.start_time is not None:
            if self.stop_time is not None:
                return self.stop_time - self.start_time
            return time.time() - self.start_time
        return None

    @property
    def enter_message(self) -> Union[str, None]:
        return self._enter_message

    @property
    def exit_message(self) -> Union[str, None]:
        return self._exit_message

    @property
    def start_time(self) -> Union[float, None]:
        return self._start_time

    @property
    def stop_time(self) -> Union[float, None]:
        return self._stop_time

    @property
    def verbose(self) -> bool:
        return self._verbose


def find_common_prefix(paths: Sequence[Union[str, Path]]) -> Optional[Path]:
    """
    Find the common prefix of two or more paths.

    ::

        >>> from pathlib import Path
        >>> one = Path('foo/bar/baz')
        >>> two = Path('foo/quux/biz')
        >>> three = Path('foo/quux/wuux')

    ::

        >>> import uqbar.io
        >>> str(uqbar.io.find_common_prefix([one, two, three]))
        'foo'

    :param paths: paths to inspect
    """
    counter: collections.Counter = collections.Counter()
    for path in paths:
        path = Path(path)
        counter.update([path])
        counter.update(path.parents)
    valid_paths = sorted(
        [path for path, count in counter.items() if count >= len(paths)],
        key=lambda x: len(x.parts),
    )
    if valid_paths:
        return valid_paths[-1]
    return None


def find_executable(name: str, flags=os.X_OK) -> List[str]:
    """
    Finds executable `name`.

    Similar to Unix ``which`` command.

    Returns list of zero or more full paths to `name`.
    """
    result = []
    extensions = [x for x in os.environ.get("PATHEXT", "").split(os.pathsep) if x]
    for directory_name in os.environ.get("PATH", "").split(os.pathsep):
        executable_path = Path(directory_name) / name
        executable_path = executable_path.resolve().absolute()
        if os.access(executable_path, flags):
            result.append(str(executable_path))
        for extension in extensions:
            extension_path = executable_path.with_suffix(extension)
            if os.access(extension_path, flags):
                result.append(str(extension_path))
    return result


def open_path(path: Path) -> None:
    if platform.system() == "Darwin":
        subprocess.run(["open", str(path)], check=True)
    elif platform.system() == "Linux":
        subprocess.run(["xdg-open", str(path)], check=True)
    elif platform.system() == "Windows":
        os.startfile(str(path))  # type: ignore


def relative_to(source_path: Union[str, Path], target_path: Union[str, Path]) -> Path:
    """
    Generates relative path from ``source_path`` to ``target_path``.

    Handles the case of paths without a common prefix.

    ::

        >>> from pathlib import Path
        >>> source = Path('foo/bar/baz')
        >>> target = Path('foo/quux/biz')

    ::

        >>> target.relative_to(source)
        Traceback (most recent call last):
          ...
        ValueError: ...

    ::

        >>> import os
        >>> import uqbar.io
        >>> str(uqbar.io.relative_to(source, target)).replace(os.path.sep, "/")
        '../../quux/biz'

    :param source_path: the source path
    :param target_path: the target path
    """
    source_path = Path(source_path).absolute()
    if source_path.is_file():
        source_path = source_path.parent
    target_path = Path(target_path).absolute()
    common_prefix = find_common_prefix([source_path, target_path])
    if not common_prefix:
        raise ValueError("No common prefix")
    source_path = source_path.relative_to(common_prefix)
    target_path = target_path.relative_to(common_prefix)
    result = Path(*[".."] * len(source_path.parts))
    return result / target_path


def walk(
    root_path: Union[str, Path], top_down: bool = True
) -> Generator[Tuple[Path, Sequence[Path], Sequence[Path]], None, None]:
    """
    Walks a directory tree.

    Like :py:func:`os.walk` but yielding instances of :py:class:`Path`
    instead of strings.

    :param root_path: foo
    :param top_down: bar
    """
    root_path = Path(root_path)
    directory_paths, file_paths = [], []
    for path in sorted(root_path.iterdir()):
        if path.is_dir():
            directory_paths.append(path)
        else:
            file_paths.append(path)
    if top_down:
        yield root_path, directory_paths, file_paths
    for directory_path in directory_paths:
        yield from walk(directory_path, top_down=top_down)
    if not top_down:
        yield root_path, directory_paths, file_paths


def write(
    contents: str, path: Union[str, Path], verbose: bool = False, logger_func=None
) -> bool:
    """
    Writes ``contents`` to ``path``.

    Checks if ``path`` already exists and only write out new contents if the
    old contents do not match.

    Creates any intermediate missing directories.

    :param contents: the file contents to write
    :param path: the path to write to
    :param verbose: whether to print output
    """
    print_func = logger_func or print
    path = Path(path)
    printed_path = str(path).replace(os.path.sep, "/")  # same display on Windows
    if path.exists():
        old_contents = path.read_text()
        if old_contents == contents:
            if verbose:
                print_func(f"preserved {printed_path}")
            return False
        else:
            path.write_text(contents)
            if verbose:
                print_func(f"rewrote {printed_path}")
            return True
    elif not path.exists():
        if not path.parent.exists():
            path.parent.mkdir(parents=True)
        path.write_text(contents)
        if verbose:
            print_func(f"wrote {printed_path}")
    return True



"""
Uqbar Sphinx executable examples extension.

Runs every Python-formatted literal block through a console, and captures
output. Can be extended to capture non-terminal output like Graphviz graphs,
other images, audio, etc. via an extension system.

The extension groups all top-level literal blocks in a document into a suite to
be executed consecutively in a console. All literal blocks in each class,
method and function docstring are grouped and executed separately in isolated
namespaces. When ``uqbar_book_use_cache`` is ``True``, the docstring grouping
allows inherited docstring suites to be executed only once per Sphinx build.

Install by adding ``'uqbar.sphinx.book'`` to the ``extensions`` list in your
Sphinx configuration.

This extension provides the following configuration values:

``uqbar_book_console_setup``:
    list of strings to run in the console before executing a suite of literal
    blocks, e.g. ``["import uqbar"]``

``uqbar_book_console_teardown``:
    list of strings to run in the console after executing a suite of literal
    blocks, even if execution halts due to error, e.g.
    ``["important_thing.shutdown()", "other_thing.cleanup()"]``

``uqbar_book_extensions``:
    list of class paths to extensions, e.g.
    ``["uqbar.book.extensions.GraphExtension"]``

``uqbar_book_strict``:
    whether to fail the Sphinx build on encountering an unexpected error

``uqbar_book_use_black``:
    whether to use ``black`` to format console input

``uqbar_book_use_cache``:
    whether to use a caching system to avoid re-executing identical literal
    block suites

"""

import importlib
from typing import Any, Dict

from docutils.nodes import SkipNode
from sphinx.util import logging
from sphinx.util.console import bold  # type: ignore

import uqbar

from ..book.console import ConsoleError
from ..book.sphinx import (
    UqbarBookDefaultsDirective,
    UqbarBookDirective,
    UqbarBookImportDirective,
    collect_literal_blocks,
    console_context,
    create_cache_db,
    group_literal_blocks_by_cache_path,
    interpret_code_blocks,
    interpret_code_blocks_with_cache,
    rebuild_document,
    uqbar_book_defaults_block,
    uqbar_book_import_block,
)

logger = logging.getLogger(__name__)


def on_builder_inited(app):
    """
    Hooks into Sphinx's ``builder-inited`` event.
    """
    app.cache_db_path = ":memory:"
    if app.config["uqbar_book_use_cache"]:
        logger.info(bold("[uqbar-book]"), nonl=True)
        logger.info(" initializing cache db")
        app.connection = create_cache_db(app.cache_db_path)


def on_config_inited(app, config):
    """
    Hooks into Sphinx's ``config-inited`` event.
    """
    extension_paths = config["uqbar_book_extensions"] or [
        "uqbar.book.extensions.GraphExtension"
    ]
    app.uqbar_book_extensions = []
    for extension_path in extension_paths:
        module_name, _, class_name = extension_path.rpartition(".")
        module = importlib.import_module(module_name)
        extension_class = getattr(module, class_name)
        extension_class.setup_sphinx(app)
        app.uqbar_book_extensions.append(extension_class)
    # Verify early that any setup / teardown works
    try:
        with console_context(
            extensions=app.uqbar_book_extensions,
            setup_lines=config["uqbar_book_console_setup"],
            teardown_lines=config["uqbar_book_console_teardown"],
        ):
            pass
    except ConsoleError:
        logger.error("uqbar.sphinx.book console setup/teardown failed")
        raise


def on_doctree_read(app, document):
    """
    Hooks into Sphinx's ``doctree-read`` event.
    """
    literal_blocks = collect_literal_blocks(document)
    cache_mapping = group_literal_blocks_by_cache_path(literal_blocks)
    node_mapping = {}
    use_cache = bool(app.config["uqbar_book_use_cache"])
    kwargs = dict(
        document=document,
        extensions=app.uqbar_book_extensions,
        setup_lines=app.config["uqbar_book_console_setup"],
        teardown_lines=app.config["uqbar_book_console_teardown"],
        use_black=bool(app.config["uqbar_book_use_black"]),
    )
    for cache_path, literal_block_groups in cache_mapping.items():
        for literal_blocks in literal_block_groups:
            try:
                if use_cache:
                    local_node_mapping = interpret_code_blocks_with_cache(
                        literal_blocks, cache_path, app.connection, **kwargs
                    )
                else:
                    local_node_mapping = interpret_code_blocks(literal_blocks, **kwargs)
                node_mapping.update(local_node_mapping)
            except ConsoleError as exception:
                message = (
                    "\n    "
                    + "\n    ".join(
                        line.rstrip()
                        for line in str(exception.args[1]).rstrip().splitlines()
                    )
                    + "\n    "
                    + "\n    ".join(
                        line.rstrip()
                        for line in exception.args[0].rstrip().splitlines()
                    )
                )
                logger.warning(message, location=exception.args[1])
                if app.config["uqbar_book_strict"]:
                    raise
    rebuild_document(document, node_mapping)


def on_build_finished(app, exception):
    """
    Hooks into Sphinx's ``build-finished`` event.
    """
    if not app.config["uqbar_book_use_cache"]:
        return
    logger.info("")
    for row in app.connection.execute("SELECT path, hits FROM cache ORDER BY path"):
        path, hits = row
        if not hits:
            continue
        logger.info(bold("[uqbar-book]"), nonl=True)
        logger.info(" Cache hits for {}: {}".format(path, hits))


def skip_node(self, node):
    raise SkipNode


def setup(app) -> Dict[str, Any]:
    """
    Sets up Sphinx extension.
    """
    app.add_config_value("uqbar_book_console_setup", [], "env")
    app.add_config_value("uqbar_book_console_teardown", [], "env")
    app.add_config_value(
        "uqbar_book_extensions", ["uqbar.book.extensions.GraphExtension"], "env"
    )
    app.add_config_value("uqbar_book_strict", False, "env")
    app.add_config_value("uqbar_book_use_black", False, "env")
    app.add_config_value("uqbar_book_use_cache", True, "env")
    app.add_config_value("uqbar_book_block_options", {}, "env")
    app.add_directive("book", UqbarBookDirective)
    app.add_directive("book-defaults", UqbarBookDefaultsDirective)
    app.add_directive("book-import", UqbarBookImportDirective)

    for node_class in [uqbar_book_defaults_block, uqbar_book_import_block]:
        app.add_node(
            node_class,
            html=[skip_node, None],
            latex=[skip_node, None],
            text=[skip_node, None],
        )
    app.connect("builder-inited", on_builder_inited)
    app.connect("config-inited", on_config_inited)
    app.connect("doctree-read", on_doctree_read)
    app.connect("build-finished", on_build_finished)
    return {
        "version": uqbar.__version__,
        "parallel_read_safe": False,
        "parallel_write_safe": True,
    }



class PublicClass:
    def __str__(self):
        return repr(self)

    def method(self):
        pass

    def other_method(self):
        pass

    def inheritable_method(self):
        pass

    @classmethod
    def class_method(cls):
        pass

    @staticmethod
    def static_method(cls):
        pass

    @property
    def read_only_property(self):
        pass

    @property
    def read_write_property(self):
        pass

    @read_write_property.setter
    def read_write_property(self, expr):
        pass


class ChildClass(PublicClass):
    def inheritable_method(self):
        pass

    def new_method(self):
        pass


class _PrivateClass:
    pass


def public_function():
    pass


def _private_function():
    pass
